/*
 * Copyright (c) 1999 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are subject to the Sun Community Source License
 * Jini(TM) Technology Core Platform, v.1.0 (the "License"); you may not use
 * this file except in compliance with the License.  You may obtain a copy of
 * the License at http://java.sun.com/products/jini. Software distributed
 * under the License is distributed on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied.  See the License for the specific
 * language governing rights and limitations under the License.
 * 
 * CopyrightVersion v1.0_Jini
 */
package com.sun.jini.collection;

/**
 * This class provides a low-synchronization list of objects.  It is
 * designed to allow proper traversal of the list while objects are
 * being added and/or removed, and to allow simultaneous adds and
 * removals.
 * <p>
 * The design is to have a list of <code>FastList.Node</code> nodes, each of
 * which has a <code>next</code> reference and a <code>removed</code>
 * flag.  New nodes are added at the tail, requiring synchronization.
 * When a node is removed from the list, its <code>removed</code> flag
 * is set to true in synchronized code.  These are two of the three
 * points of synchronization (see below for the third).
 * <p>
 * Nodes are not directly removed from the list.  When a reader of the
 * list traverses the list, invoking <code>FastList.Node.next</code>,
 * removed nodes are skipped over, and the node's <code>next</code>
 * reference is updated to skip over them.  Thus, removed nodes are
 * progressively orphaned as the node connects around them, leaving them
 * unreachable, and hence garbage.  This practice can be aided by a global
 * reaper thread, which goes through each list intermittently, reaping
 * removed nodes.  This prevents a list whose nodes are removed from
 * the end from having a large number of trailing removed nodes.  When
 * the head of the list is fetched, if it has been removed, the head is
 * updated to point to the next unremoved list (or <code>null</code>).
 * This is the third point of synchronization -- fetching the head when
 * the head node as been removed.  Global reaping can be turned off or
 * on -- it is off by default.
 * <p>
 * All this means that users of the list will occasionally see a
 * removed node in the list, because it may be marked removed after it
 * has been reached while traversing the list.  This will be rare, but
 * it is up to the user of the list to skip over nodes that have been
 * removed if necessary, and to handle or avoid attempted re-removal of
 * a node.  This is why <code>remove</code> returns a boolean -- if the
 * node has already been removed by the time the attempt to remove it
 * happens, <code>remove</code> will return <code>false</code>.  This
 * is done synchronizing on the node itself.  A typical piece of code
 * to remove would look like this:
 *
 * <pre>
 *  public synchronized Value findMatch(Value matchFor) {
 *	for (Value val = list.head(); val != null; val = val.next()) {
 *	    if (matchFor.matches(val)) {
 *		if (!list.remove(n))	// oh, well -- someone got to it first
 *		    continue;		// keep looking
 *		return val;		// it's ours to take!
 *	    }
 *	}
 *	return null;
 *  }
 * </pre>
 *
 * This code assumes that getting a removed node will be rare enough
 * that there is no point in testing for it until it's found a match.
 * If testing for a match was expensive, one might
 * <code>synchronize</code> on the node object and testing for it being
 * removed and then trying for matching, thus preventing any other
 * thread from attempting a match at the same time.  In any case, the
 * <code>remove</code> test is itself atomic -- either it succeeds, or
 * it fails because someone got there first.
 * <p>
 * The reaping thread runs at <code>Thread.MIN_PRIORITY,</code> and
 * simply iterates over the list to force the <code>next</code>
 * pointers to be updated.  The pause time between reaping runs is
 * controllable.  The default is to sleep 60 seconds between reap
 * attempts, although a reaping run can be forced by invoking the
 * static <code>reapAll</code> method.  You can reap an indvidual thread
 * by invoking it's <code>reap</code> method.
 *
 * @author Ken Arnold
 */

// NOTE: if you change the implementation, remember that any
// modificaiton of the head reference requires synchronization

public class FastList implements Runnable {

    /**
     * A single node in a <code>FastList</code>.
     */
    public static class Node {
	public  Node	next;
	private Node	prev;
	private boolean	removed;	// has this node been removed?

	/**
	 * Return the next node, skipping over removed ones.  Note that
	 * between the time that the next node is returned and the time you
	 * look at it someone may have removed it.
	 */
	public final Node next() {
	    if (FastList.doDebugYield) FastList.debugYield();//DEBUG
	    Node n = next;
	    if (FastList.doDebugYield) FastList.debugYield();//DEBUG

	    // We have to be careful to always leave the tail node, even
	    // if it is removed, since new nodes are added to the tail
	    while (n != null && n.removed && n.next != null) {
	        if (FastList.doDebugYield) FastList.debugYield();//DEBUG
		n = n.next;
	    }
	    if (FastList.doDebugYield) FastList.debugYield();//DEBUG
	    return (next = n);	// update next to orphan any removed node(s)
	    /*
	     * The assignment in the return statement may unnecessarily
	     * dirty the page if n == next.  Is this a problem?  We could
	     * just not bother, and let the reaper in FastList do all the
	     * work, but this seems worth the tradeoff.
	     */
	}

	/**
	 * Return the prev node, skipping over removed ones.  Note that
	 * between the time that the prev node is returned and the time you
	 * look at it someone may have removed it.
	 */
	public final Node prev() {
	    if (FastList.doDebugYield) FastList.debugYield();//DEBUG
	    Node p = prev;
	    if (FastList.doDebugYield) FastList.debugYield();//DEBUG
	    while (p != null && p.removed) {
	        if (FastList.doDebugYield) FastList.debugYield();//DEBUG
		p = p.prev;
	    }
	    if (FastList.doDebugYield) FastList.debugYield();//DEBUG
	    return (prev = p);	// update next to orphan any removed node(s)
	    /*
	     * The assignment in the return statement may unnecessarily
	     * dirty the page if p == prev.  Is this a problem?  We could
	     * just not bother, and let the reaper in FastList do all the
	     * work, but this seems worth the tradeoff.
	     */
	}

	/**
	 * Return <code>true</code> if the object has been removed from the list
	 */
	public final boolean removed() {
	    if (FastList.doDebugYield) FastList.debugYield();//DEBUG
	    return removed;
	}

	/**
	 * Clear the <code>removed</code> flag.
	 */
	private final void unremove() {
	    if (FastList.doDebugYield) FastList.debugYield();//DEBUG
	    removed = false;
	}

	/**
	 * Set the <code>next</code> reference for this node.
	 */
	private final void next(Node n) {
	    if (FastList.doDebugYield) FastList.debugYield();//DEBUG
	    next = n;
	}

	/**
	 * Set the <code>prev</code> reference for this node.
	 */
	private final void prev(Node p) {
	    if (FastList.doDebugYield) FastList.debugYield();//DEBUG
	    prev = p;
	}

	/**
	 * Remove the node from the list.  This will be unsuccessful if the
	 * node was removed by someone else after you got your hands on
	 * it.  This must be synchronized because the test to check if it
	 * is already removed and the following decision to remove it if it
	 * hasn't been removed must be unitary.
	 */
	private synchronized final boolean remove() {
	    if (FastList.doDebugYield) FastList.debugYield();//DEBUG
	    if (removed)
		return false;
	    if (FastList.doDebugYield) FastList.debugYield();//DEBUG
	    return (removed = true);
	}

	final void dump(java.io.PrintWriter out) { //DEBUG
	  if (removed) //DEBUG
	      out.print('!'); //DEBUG
	  out.print(toString()); //DEBUG
	  if (next != null) { //DEBUG
	      out.print(' '); //DEBUG
	      next.dump(out); //DEBUG
	  } //DEBUG
	} //DEBUG
    }

    private Node	head;			// the head of the list
    private Node	tail;			// the tail of the list
    private boolean	reapable;		// modified since last reap?

    private static FastList 	lists;			// list of all FastLists
    private static Thread	reaper;			// thread for the reaper
    private static boolean	stopReaping = false;	// interrupt means stop
    private static boolean	reapSomeone = false;	// a list needs reaping?

    private static int		pauseTime = 60000;	// every 60 seconds

    static final boolean	doDebugYield = false;	// see TestFastList

    /*
     * Set up the list of lists and the reaper thread to browse it.
     */
    static {
	lists = new FastList(false);
	globalReap(false);
	//DEBUG System.out.println("reaper is started");
    }

    /**
     * Create an empty list without adding it to the list of lists.
     * This is private because it is the constructor used to
     * <i>create</i> the list of lists to avoid the chicken-and-egg
     * problem.  The <code>boolean</code> is there only to allow an
     * overlead of the constructor.
     */
    private FastList(boolean addTo) {
    }

    /**
     * Create an new empty list.
     */
    public FastList() {
	lists.add(new FastListNode());
    }

    // the outer this is the only interesting data in this class
    private class FastListNode extends Node {
	void reap() { FastList.this.reap(); }
    }

    /**
     * Add a given node to the end of the list.
     */
    public void add(Node node) {
	if (doDebugYield) debugYield();//DEBUG
	synchronized (this) {
	    if (tail == null) {
		tail = head = node;
		node.next(null);
		node.prev(null);
	    }
	    else {
		node.next(null);
		synchronized (tail) {
		    tail.next(node);
		    node.prev(tail);
		    tail = node;
		}
	    }
	}
	if (doDebugYield) debugYield();//DEBUG
    }

    /**
     * Put a previously-removed node back into the list.
     */
    public void replace(Node node) {
	if (doDebugYield) debugYield();//DEBUG
	node.unremove();
	if (doDebugYield) debugYield();//DEBUG
	synchronized (this) {
	    add(node);
	}
	if (doDebugYield) debugYield();//DEBUG
    }

    /**
     * Add a given node to the list after the specified node.  The
     * specified node must not be removed, or an
     * <code>IllegalArgumentException</code> is thrown.  If
     * <code>after</code> is <code>null</code>, <code>addAfter</code>
     * is equivalent to <code>add</code>.
     */
    public void addAfter(Node after, Node node) {
	if (after == null) {
	    // no need for debug code here -- add() has its own
	    add(node);
	    return;
	}

	/*
	 * We must synchronize so that the node "after" isn't orphaned while
	 * we adding this.
	 */
	if (doDebugYield) debugYield();//DEBUG
	synchronized (after) {
	    if (doDebugYield) debugYield();//DEBUG
	    if (after.removed()) {
		throw new IllegalArgumentException(
		    "Adding node after a removed node");
	    }
	    if (doDebugYield) debugYield();//DEBUG
	    synchronized (node) {
		node.next(after.next());
		node.prev(after);
		if (node.next() != null)
		    node.next().prev(node);
	    }
	    after.next(node);
	    if (doDebugYield) debugYield();//DEBUG
	}

	synchronized(this) {
	    if (tail == after)
	        tail = node;
	}

	if (doDebugYield) debugYield();//DEBUG
    }

    /**
     * Return the start of the list.  This is unsynchronized unless the
     * existing head has been removed, when it must then synchronize on
     * the list to update the head.
     */
    // We could leave this optimization to the reap, but this seems better
    public Node head() {
	if (doDebugYield) debugYield();//DEBUG
	if (head != null && head.removed()) {
	    synchronized (this) {
		// retest -- it may have changed between the test and here
		if (head != null && head.removed())
		    head = head.next();
		if (head == null)
		    tail = null;
	    }
	}
	return head;
    }

    /**
     * Return the tail of the list.
     */
    // we could leave this optimization to the reap, but this seems better
    public Node tail() {
	if (doDebugYield) debugYield();//DEBUG
	if (tail != null) {
	    synchronized (this) {
		// retest -- it may have changed between the test and here
		if (tail == null)
		    return null;
		synchronized (tail) {
		    if (tail != null && tail.removed())
			tail = tail.prev();
		    if (tail == null)
			head = null;
		}
	    }
	}
	return tail;
    }

    /**
     * Remove the given node from this list.  This requires
     * synchronization on the node object.
     */
    public boolean remove(Node node) {
						if (doDebugYield) debugYield();
	if (!node.remove()) {			if (doDebugYield) debugYield();
	    return false;
	}
	reapable = true;			if (doDebugYield) debugYield();
	reapSomeone = true;			if (doDebugYield) debugYield();
	return true;
    }

    /**
     * Return an enumeration over the contents of the list.
     */
    public java.util.Enumeration objs() {
	return new java.util.Enumeration() {
	    Node cur = head();

	    public boolean hasMoreElements() {
		return (cur.next() != null);
	    }

	    public Object nextElement() {
		Node next = cur.next();
		if (next == null)
		    throw new java.util.NoSuchElementException();
		return next;
	    }
	};
    }

    /**
     * Reap the removed nodes of this list.  You should need this rarely,
     * if ever.
     *
     * @see #reapAll
     */
    public void reap() {
	if (doDebugYield) debugYield();//DEBUG
	if (reapable) {	// does this list need reaping?
	    reapCnt++;//DEBUG
	    if (doDebugYield) debugYield();//DEBUG
	    reapable = false;	// not any more
	    if (doDebugYield) debugYield();//DEBUG
	    // iteration happens to reap, so that's all we must do
	    if (doDebugYield) debugYield();//DEBUG
	    synchronized (this) {
		if (head == null)
		    return;
		if (head.removed())
		    head = head.next();
		if (head == null)
		    tail = null;
		else if (tail.removed())
		    tail = tail.prev();
	    }
	    if (doDebugYield) debugYield();//DEBUG
	    for (Node n = head; n != null; n = n.next()) {
		if (doDebugYield) debugYield();//DEBUG
		continue;
	    }
	    for (Node n = tail; n != null; n = n.prev()) {
		if (doDebugYield) debugYield();//DEBUG
		continue;
	    }
	}
    }

    int reapCnt = 0; //DEBUG
    public final int reapCnt() { return reapCnt; } //DEBUG

    /**
     * Reap all known lists right now.  This should be rarely, if ever,
     * needed.
     *
     * @see #reap
     */
    /*
     * A list needs to be reaped only if something has been removed
     * from it.  Since <code>Node.next</code> already orphans
     * removed nodes when encountered, reaping a list simply requires
     * iterating over it.  We might even find this superfluous, since
     * any traversal of the list searching for things will similarly
     * reap the list, but we cannot be sure that any given removed node
     * will happen to be crossed by searches.  However, we should
     * probably make the wait time get longer if, in a given server,
     * runs that actually <b>do</b> reap something are rare because the
     * particular access patterns to that server cause reaps to happen
     * in the normal course of events.
     */
    //!! Adjust pause times as described above
    //!! should use weak references
    public static void reapAll() {
	if (doDebugYield) debugYield();//DEBUG
	if (reapSomeone) {		// does *anyone* need reaping?
	    //System.out.println("reaping");
	    if (doDebugYield) debugYield();//DEBUG
	    reapSomeone = false;	// not any more
	    if (doDebugYield) debugYield();//DEBUG
	    for (FastListNode list = (FastListNode) lists.head();
		 list != null;
		 list = (FastListNode) list.next())
	    {
		if (doDebugYield) debugYield();//DEBUG
		list.reap();
	    }
	} else {
	    //System.out.println("nothing to reap");
	}
	if (doDebugYield) debugYield();//DEBUG
    }

    /**
     * Return whether or not intermittent global reaping is to be done.
     *
     * @see #globalReap(boolean)
     * @see #pauseTime()
     * @see #pauseTime(int)
     */
    public static synchronized boolean globalReap() {
	return (reaper != null);
    }

    /**
     * Set whether or not intermittent global reaping is to be done.
     *
     * @see #globalReap()
     * @see #pauseTime()
     * @see #pauseTime(int)
     */
    public static synchronized void globalReap(boolean doIt) {
	if (doIt == globalReap())
	    return;
	if (doIt) {
	    stopReaping = false;
	    reaper = new Thread(lists);
	    reaper.setPriority(Thread.MIN_PRIORITY);
	    reaper.setDaemon(true);
	    reaper.start();
	} else {
	    stopReaping = true;
	    reaper.interrupt();
	    reaper = null;
	}
    }

    /**
     * Return the current pauseTime setting for the reaper.
     *
     * @see #pauseTime(int)
     * @see #globalReap()
     * @see #globalReap(boolean)
     */
    public static int pauseTime() {
	return pauseTime;
    }

    /**
     * Set the pause time for the reaper.  This will take affect on the
     * next iteration.
     *
     * @see #reapAll
     * @see #pauseTime()
     * @see #globalReap()
     * @see #globalReap(boolean)
     */
    public static void pauseTime(int time) {
	pauseTime = time;
	reaper.interrupt();	// if it's sleeping, try the new time
    }

    /**
     * Set the priority of the reaping thread.
     */
    public static void reapPriority(int p) {
	reaper.setPriority(p);
    }

    /**
     * Should this debug run yeild the thread?
     */
    static void debugYield() {
	// if (TestFastList.randBoolean())//DEBUG
	    Thread.yield();//DEBUG
    }

    public void dump(java.io.PrintWriter out) {
	if (head == null) //DEBUG
	    out.print("null"); //DEBUG
	else //DEBUG
	    head.dump(out); //DEBUG
    }

    /**
     * This <code>run</code> method is used by the reaper list to force
     * the orphaning of removed nodes in lists.  It should <b>never</b>
     * be used on any other list.  It is only <code>public</code>
     * because the <code>Runnable</code> interface requires that it be
     * so.  It is really an internal implementation detail of
     * <code>FastList</code>.
     */
    // create a separate class to hold this away from the public contract?
    public void run() {
	if (this != lists)
	    throw new IllegalArgumentException("this must be 'lists' in run()");

	for (;;) {
	    try {
						if (doDebugYield) debugYield();
		Thread.sleep(pauseTime);	if (doDebugYield) debugYield();
		reapAll();			if (doDebugYield) debugYield();
	    } catch (InterruptedException e) {
		System.out.println("interrupted");//DEBUG
		if (stopReaping)
		    return;
		;	// otherwise must mean pause time has changed
	    }
	}
    }
}
