/*
 * Copyright (c) 1999 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are subject to the Sun Community Source License
 * Jini(TM) Technology Core Platform, v.1.0 (the "License"); you may not use
 * this file except in compliance with the License.  You may obtain a copy of
 * the License at http://java.sun.com/products/jini. Software distributed
 * under the License is distributed on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied.  See the License for the specific
 * language governing rights and limitations under the License.
 * 
 * CopyrightVersion v1.0_Jini
 */
package com.sun.jini.lease.landlord;

import java.io.Serializable; 
import java.rmi.Remote;
import java.rmi.RemoteException;
import net.jini.core.lease.Lease;
import net.jini.core.lease.LeaseMapException;
import net.jini.core.lease.LeaseDeniedException;
import net.jini.core.lease.UnknownLeaseException;

/** 
 * Interface that owners of leased resources must meet inorder to
 * work with the <code>jini.lease.Lease</code>s returned by
 * <code>LeaseDurationPolicy.leaseFor()</code>.
 * @see net.jini.core.lease.Lease
 * @see LeaseDurationPolicy
 */
public interface Landlord extends Remote {
    /**
     * Called by the lease when its <code>renew</code> method is called.
     *
     * @param cookie Associated with the lease when it was created
     * @param extension
     *               The duration argument passed to the
     *               <code>Lease.renew()</code> call
     * @return The new duration the lease should have
     * @see LeasePolicy#leaseFor
     */
    public long renew(Object cookie, long extension)
	throws LeaseDeniedException, UnknownLeaseException, RemoteException;
       
    /**
     * Called by the lease when its <code>cancel</code> method is called.
     *
     * @param cookie Associated with the lease when it was created
     * @see LeasePolicy#leaseFor
     */
    public void cancel(Object cookie)
	throws UnknownLeaseException, RemoteException;

    /**
     * Called by the lease map when its <code>renewAll</code> method is called.
     * @param cookie Associated with each lease when it was created
     * @param extension
     *               The duration argument for each lease from the map
     * @return The results of the renew
     * @see LeasePolicy#leaseFor
     */
    public RenewResults renewAll(Object[] cookie, long[] extension)
	throws RemoteException;
       
    /**
     * Called by the lease map when its <code>cancelAll</code> method is called.
     * @param cookie Associated with the lease when it was created
     * @see LeasePolicy#leaseFor
     */
    public void cancelAll(Object[] cookie)
	throws LeaseMapException, RemoteException;


    /** 
     * Simple class that holds return values of
     * the <code>Landlord.renewAll</code> method.
     *
     * @see Landlord#renewAll
     */
    public class RenewResults implements java.io.Serializable {
	static final long serialVersionUID = 5053579718775268824L;

	/**
	 * For each cookie passed to <code>renewAll</code>,
	 * <code>granted[i]</code> is the granted lease time, or -1 if the
	 * renewal for that least generated an exception.  If there was
	 * an exception, the exception is held in <code>denied</code>.
	 *
	 * @see #denied
	 * @serial
	 */
	public long[] granted;

	/**
	 * The <code>i</code><sup><i>th</i></sup> -1 in <code>granted</code>
	 * was denied because of <code>denied[i]</code>.  If nothing was denied,
	 * this field is <code>null</code>.
	 *
	 * @serial
	 */
	public Exception[] denied;

	/**
	 * Create a <code>RenewResults</code> object setting the field
	 * <code>granted</code> to the passed value, and <code>denied</code>
	 * to <code>null</code>.
	 *
	 * @param granted	The value for the field <code>granted</code>
	 */
	public RenewResults(long[] granted) {
	    this(granted, null);
	}

	/**
	 * Create a <code>RenewResults</code> object setting the field
	 * <code>granted</code> and <code>denied</code> fields to the
	 * passed values.
	 *
	 * @param granted	The value for the field <code>granted</code>
	 * @param denied	The value for the field <code>denied</code>
	 */
	public RenewResults(long[] granted, Exception[] denied) {
	    this.granted = granted;
	    this.denied = denied;
	}
    };
}
