/*
 * Copyright (c) 1999 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are subject to the Sun Community Source License
 * Jini(TM) Technology Core Platform, v.1.0 (the "License"); you may not use
 * this file except in compliance with the License.  You may obtain a copy of
 * the License at http://java.sun.com/products/jini. Software distributed
 * under the License is distributed on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied.  See the License for the specific
 * language governing rights and limitations under the License.
 * 
 * CopyrightVersion v1.0_Jini
 */
package com.sun.jini.lease.landlord;

import java.io.Serializable;
import java.rmi.RemoteException;

import net.jini.core.lease.Lease;
import net.jini.core.lease.LeaseMap;
import net.jini.core.lease.LeaseDeniedException;
import net.jini.core.lease.UnknownLeaseException;
import com.sun.jini.lease.AbstractLease;

/**
 * Basic implementaion of <code>net.jini.core.lease.Lease</code> that works
 * with the the Landlord protocol.
 *
 * @see Landlord
 */
public class LandlordLease extends AbstractLease {
    static final long serialVersionUID = 597539342757316824L;

    /**
     * Object passed to <code>landlord</code> when renewal and cancel
     * calls are make on this lease.  Should uniquely identify the lease.
     * @serial 
     */
    final private Object   cookie;

    /**
     * Owner of the resource associated with this lease.
     * @serial
     */
    final private Landlord landlord;

    /**
     * Create a new lease.
     * @param landlord <code>Landlord</code> object that will be used to
     *                 communicate renew and cancel requests to the granter
     *                 of the lease.
     * @param cookie   An object that the <code>Landlord</code> can use
     *                 to identify the leased resource.
     * @param duration The initial expiration of the lease.
     */
    LandlordLease(Object cookie, Landlord landlord, long expiration) {
	super(expiration);
	this.cookie   = cookie;
	this.landlord = landlord;
    }

    // Implementation of the Lease interface

    // purposefully inherit doc comment from supertype
    public void cancel() throws UnknownLeaseException, RemoteException {
	landlord.cancel(cookie);
    }

    // purposefully inherit doc comment from supertype
    protected long doRenew(long renewDuration)
	 throws LeaseDeniedException, UnknownLeaseException, RemoteException
    {
	if (renewDuration < 0 && !(renewDuration == Lease.FOREVER ||
				   renewDuration == Lease.ANY))
	{
	    throw new IllegalArgumentException("Lease renewal: " +
	        "Asked for a negative duration");
	}
	return landlord.renew(cookie, renewDuration);
    }

    // purposefully inherit doc comment from supertype
    public boolean equals(Object other) {
	// Note, we do not include the expiration in the equality test.
	// If the lease is copied and ether the copy or the original
	// is renewed they are conceptually the same because they
	// still represent the same claim on the same resource
	// --however their expiration will be different

	if (other instanceof LandlordLease) {
	    LandlordLease that = (LandlordLease)other;
	    return cookie.equals(that.cookie) &&
		landlord.equals(that.landlord);
	}

	return false;
    }

    // inherit doc comment
    public boolean canBatch(Lease lease) {
	if (lease instanceof LandlordLease)
	    return landlord.equals(((LandlordLease) lease).landlord);
	return false;
    }

    /** Return the landlord. */
    Landlord landlord() {
	return landlord;
    }

    /** Return the cookie. */
    Object cookie() {
	return cookie;
    }

    // inherit doc comment
    public LeaseMap createLeaseMap(long duration) {
	return new LandlordLeaseMap(landlord, this, duration);
    }

    // purposefully inherit doc comment from supertype
    public int hashCode() {
	return (int) cookie.hashCode() ^ landlord.hashCode();
    }

    // purposefully inherit doc comment from supertype
    public String toString() {
	return "LandlordLease:" + cookie + " landlord:" + landlord +
		super.toString();
    }

    /** Set the expiration. */
    void setExpiration(long expiration) {
	this.expiration = expiration;
    }

    /**
     * Implementation of <code>LandlordLeaseFactory</code> the produces
     * <code>LandlordLease</code>s.
     */
    public static class Factory implements LandlordLeaseFactory {
	public Lease newLease(Object cookie, Landlord landlord, long expiration)
	{
	    // purposefully inherit doc comment from supertype
	    return new LandlordLease(cookie, landlord, expiration);
	}
    }

}
