/*
 * Copyright (c) 1999 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are subject to the Sun Community Source License
 * Jini(TM) Technology Core Platform, v.1.0 (the "License"); you may not use
 * this file except in compliance with the License.  You may obtain a copy of
 * the License at http://java.sun.com/products/jini. Software distributed
 * under the License is distributed on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied.  See the License for the specific
 * language governing rights and limitations under the License.
 * 
 * CopyrightVersion v1.0_Jini
 */
package com.sun.jini.lease.landlord;

import net.jini.core.lease.*;
import com.sun.jini.lease.AbstractLeaseMap;

import java.rmi.RemoteException;
import java.util.Map;
import java.util.Iterator;
import java.util.HashMap;
import java.util.ConcurrentModificationException;

/**
 * Implementaion of <code>LeaseMap</code> for <code>LandlordLease</code>.
 *
 * @see LandlordLease
 * @see net.jini.core.lease.LeaseMap
 */
public class LandlordLeaseMap extends AbstractLeaseMap {
    static final long serialVersionUID = -5754145801478146770L;

    /**
     * The landlord which this map will talk to.
     *
     * @serial
     */
    private Landlord landlord;

    /**
     * Create a new <code>LandlordLeaseMap</code>.
     * @param landlord Owner of the resource <code>lease</code> is for.
     * @param lease    First lease to be placed in the map.  It is
     *                 assumed that <code>canContainKey(lease)</code>
     *                 is <code>true</code>.  Must work with the 
     *                 landlord protocol.
     * @param duration The duration the lease should be renewed for if 
     *                 <code>renewAll</code> is called.
     */
    LandlordLeaseMap(Landlord landlord, Lease lease, long duration) {
	super(lease, duration);
	this.landlord = landlord;
    }

    // inherit doc comment
    public boolean canContainKey(Object key) {
	if (key instanceof LandlordLease)
	    return landlord.equals(((LandlordLease) key).landlord());
	else
	    System.out.println("map can't contain " + key.getClass().getName());
	return false;
    }

    // inherit doc comment
    public void cancelAll() throws LeaseMapException, RemoteException {
	Object[] cookies = new Object[size()];
	LandlordLease[] leases = new LandlordLease[cookies.length];
	Iterator it = keySet().iterator();
	for (int i = 0; it.hasNext(); i++) {
	    LandlordLease lease = (LandlordLease) it.next();
	    leases[i] = lease;
	    cookies[i] = lease.cookie();
	}

	try {
	    landlord.cancelAll(cookies);
	} catch (LeaseMapException e) {
	    e.printStackTrace();
	    // translate the cookie->exception map into a lease->exception map
	    Map map = e.exceptionMap;
	    int origSize = map.size();
	    for (int i = 0; i < cookies.length; i++) {
		Object exception = map.remove(cookies[i]); // harmless if no map
		if (exception != null) {		   // if there was a map
		    map.put(leases[i], exception);	   // put back as lease
		    remove(leases[i]);			   // remove from map
		}
	    }
	    if (origSize != map.size())		// some cookie wasn't found
		throw new ConcurrentModificationException();
	    throw e;
	}
    }

    // inherit doc comment
    public void renewAll() throws LeaseMapException, RemoteException {
	Object[] cookies = new Object[size()];
	long[] extensions = new long[cookies.length];
	LandlordLease[] leases = new LandlordLease[cookies.length];
	Iterator it = keySet().iterator();
	for (int i = 0; it.hasNext(); i++) {
	    LandlordLease lease = (LandlordLease) it.next();
	    leases[i] = lease;
	    cookies[i] = lease.cookie();
	    extensions[i] = ((Long) get(lease)).longValue();
	}

	long now = System.currentTimeMillis();
	Landlord.RenewResults results = landlord.renewAll(cookies, extensions);

	Map bad = null;
	int d = 0;
	for (int i = 0; i < cookies.length; i++) {
	    if (results.granted[i] != -1)
		leases[i].setExpiration(now + results.granted[i]);
	    else {
		if (bad == null) {
		    bad = new HashMap(results.denied.length +
				      results.denied.length / 2);
		}
		Object badTime = remove(leases[i]);	// remove from this map
		if (badTime == null)			// better be in there
		    throw new ConcurrentModificationException();
		bad.put(leases[i], results.denied[d++]);// add to "bad" map
	    }
	}

	if (bad != null)
	    throw new LeaseMapException("renewing", bad);
    }
}
