/*
 * Copyright (c) 1999 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are subject to the Sun Community Source License
 * Jini(TM) Technology Core Platform, v.1.0 (the "License"); you may not use
 * this file except in compliance with the License.  You may obtain a copy of
 * the License at http://java.sun.com/products/jini. Software distributed
 * under the License is distributed on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied.  See the License for the specific
 * language governing rights and limitations under the License.
 * 
 * CopyrightVersion v1.0_Jini
 */
package com.sun.jini.lookup;

import net.jini.core.lease.*;
import net.jini.core.entry.Entry;
import net.jini.core.discovery.LookupLocator;
import net.jini.discovery.*;
import net.jini.core.lookup.*;
import com.sun.jini.lease.*;
import com.sun.jini.discovery.*;
import com.sun.jini.lookup.entry.*;

import java.io.IOException;
import java.lang.reflect.*;
import java.util.*;
import java.rmi.MarshalledObject;
import java.rmi.RemoteException;


/**
* This class manages the join protocol for a service.  It discovers and
* keeps track of which lookup services to join, registers with them, keeps
* the registration leases renewed, and keeps the attributes up to date.
*/

public class JoinManager{
    final private static int ADD_ATTRIBUTES = 0;
    final private static int MODIFY_ATTRIBUTES = 1;
    final private static int SET_ATTRIBUTES = 2;

    final private static int FROM_GROUP = 1;
    final private static int FROM_LOCATOR = 2;
    
    private ServiceIDListener callback;


    private ServiceItem serviceItem;
    private ArrayList joinSet = new ArrayList();
    private TaskThread taskThread = null;
    private ArrayList task = new ArrayList();

    private LeaseRenewalManager leaseRenewalMgr;
    private LookupLocatorDiscovery locatorDisc;
    private LookupDiscovery lookupDisc;
    private DiscMgrListener locatorlistener = 
                            new DiscMgrListener(FROM_LOCATOR);

    private DiscMgrListener grouplistener = 
                            new DiscMgrListener(FROM_GROUP);
    private boolean bTerminated = false;
    private class DiscMgrListener implements DiscoveryListener {
	private int from;
	public DiscMgrListener(int from) {
	    this.from = from;
	}

	public void discovered(DiscoveryEvent e) {
	    //System.out.println("discovered..." + from);
	    ServiceRegistrar[] proxys = (ServiceRegistrar[])e.getRegistrars();
	    RegisterTask t = new RegisterTask(proxys, from);
	    createTaskThread(t);

	}

	public void discarded(DiscoveryEvent e) {
	    //System.out.println("remove...");
	    ServiceRegistrar[] proxys = (ServiceRegistrar[])e.getRegistrars();
	    DiscardProxyTask t = 
		new DiscardProxyTask(proxys,from);
	    createTaskThread(t);
	}
	
    }

   /**
     * Use this constructor when the service does not yet have a service ID,
     * and has not been specially pre-configured with respect to joins.
     * Initially there are no specific lookup services to join, and groups
     * is set to {""}.
     *
     * @param obj the service object
     * @param attrSets the attribute sets for the service
     * @param callback gets passed the assigned service ID so that the
     * service can make it persistent.
     *
     * @see ServiceIDListener
     */
    public JoinManager(Object obj,  Entry[] attrSets,
		       ServiceIDListener callback,
		       LeaseRenewalManager leaseMgr)
	throws IOException
    {
	createJoinManager(null, obj, attrSets, new String[]{""},
			  null, callback, leaseMgr);
    }


    /** 
     * Use this constructor when the service does not have a service ID,
     * but has been specially pre-configured with respect to joins.
     *
     * @param obj the service object
     * @param attrSets the attribute sets for the service
     * @param groups groups to join.  Null (or empty) means join no groups,
     * not join all groups.
     * @param locators locators of specific lookup services to join
     * @param callback gets called with the assigned service ID so that
     * the service can make it persistent
     *
     * @see ServiceIDListener
     */
    public JoinManager(Object obj, Entry[] attrSets, String[] groups,
		       LookupLocator[] locators,
		       ServiceIDListener callback,
		       LeaseRenewalManager leaseMgr )
	throws IOException
    {
	createJoinManager(null, obj, attrSets, groups,
			  locators, callback, leaseMgr);	
    }

    /**
     * Use this constructor when the service already has a service ID.
     *
     * @param serviceID the service ID (must not be null)
     * @param obj the service object
     * @param attrSets the attribute sets for the service
     * @param groups groups to join.  Null (or empty) means join no groups,
     * not join all groups.
     * @param locators locators of specific lookup services to join
     */
    public JoinManager(ServiceID serviceID, Object obj,  Entry[] attrSets, 
		       String[] groups, LookupLocator[] locators,
		       LeaseRenewalManager leaseMgr) 
	throws IOException
    {
	createJoinManager(serviceID, obj, attrSets, 
			    groups, locators, callback, leaseMgr);
    }

    private void createJoinManager(ServiceID serviceID, 
				Object obj,Entry[] attrSets, 
				String[] groups, LookupLocator[] locators,
				ServiceIDListener callback,
				LeaseRenewalManager leaseMgr) 
	throws IOException
    {
	Entry[] lookupAttr = attrSets;
	if(lookupAttr == null)
	    lookupAttr = new Entry[0];

	serviceItem = new ServiceItem(serviceID, obj, lookupAttr);
	leaseRenewalMgr = ( leaseMgr != null) ? leaseMgr :
	                  new LeaseRenewalManager();

	this.callback = callback;
	lookupDisc = new LookupDiscovery(groups);
	lookupDisc.addDiscoveryListener(grouplistener);
	locatorDisc = new LookupLocatorDiscovery(locators);
	locatorDisc.addDiscoveryListener(locatorlistener);
    }
 
    /**
     * Get the list of lookup services that have currently been joined.
     * A new array is returned each time.
     */
    public ServiceRegistrar[] getJoinSet() {
	synchronized(joinSet) {
	    ServiceRegistrar[] ret = new ServiceRegistrar[joinSet.size()];
	    int k = 0;
	    for (Iterator iter = joinSet.iterator(); iter.hasNext(); ) {
		ret[k++] = ((ProxyReg)iter.next()).proxy;
	    }
	    return ret;
	}
    }

    /**
     * Get the list of locators of specific lookup services to join.
     * A new array is returned each time.
     */
    public LookupLocator[] getLocators(){
	return locatorDisc.getLocators();
    }

    

    /**
     * Add new locators for specific lookup services to join.
     *
     * @param locators locators of specific lookup services to join
     */
    public void addLocators(LookupLocator[] locators){
	locatorDisc.addLocators(locators);
    }

    /**
     * Remove locators for specific lookup services from the set to join.
     * Any leases held at the lookup services are cancelled.
     *
     * @param locators locators of specific lookup services to leave
     */
    public void removeLocators(LookupLocator[] locators) {
	locatorDisc.removeLocators(locators);
    }

    /**
     * Replace the list of locators of specific lookup services to join
     * with a new list.  Leases are cancelled at lookup services that were
     * in the old list but are not in the new list.
     *
     * @param locators locators of specific lookup services to join
     */
    public void setLocators(LookupLocator[] locators) {
	locatorDisc.setLocators(locators);

    }
    
    /** Get the list of groups to join.  A new array is returned each time. */
    public String[] getGroups() {
	return lookupDisc.getGroups();
    }

    /**
     * Add new groups to to the set to join.
     *
     * @param groups groups to join
     * @throws IOException if cannot create the MulticastSocket
     */
    public void addGroups(String[] groups) throws IOException {
	lookupDisc.addGroups(groups);

    }


    /**
     * Remove groups from the set to join.  Leases are cancelled at lookup
     * services that are not members of any of the remaining groups.
     *
     * @param groups groups to leave
     */
    public void removeGroups(String[] groups) throws IOException {
	lookupDisc.removeGroups(groups);
    }


    /**
     * Replace the list of groups to join with a new list.  Leases are
     * cancelled at lookup services that are not members of any of the
     * new groups.
     *
     * @param groups groups to join
     * @throws IOException if cannot create the MulticastSocket
     */
    public void setGroups(String[] groups) throws IOException {
	lookupDisc.setGroups(groups);
    }



    /** Get the current attribute sets.  A new array is returned each time. */
    public Entry[] getAttributes() {
	return (Entry[])serviceItem.attributeSets.clone();
    }

    /** 
     * Add attribute sets for the service.  The resulting set will be used
     * for all future joins.  The attribute sets are also added at all
     * currently joined lookup services.
     *
     * @param attrSets the attribute sets to add
     */
    public void addAttributes(Entry[] attrSets) {
	addAttributes(attrSets, false);
    }


    /**  
     * Modify the current attribute sets, using the same semantics as
     * ServiceRegistration.modifyAttributes.  The resulting set will be used
     * for all future joins.  The same modifications are also made at all
     * currently joined lookup services.
     *
     * @param attrSetTemplates the templates for matching attribute sets
     * @param attrSets the modifications to make to matching sets
     *
     * @see net.jini.core.lookup.ServiceRegistration#modifyAttributes
     */
    public void modifyAttributes(Entry[] attrSetTemplates, Entry[] attrSets) {
	modifyAttributes(attrSetTemplates, attrSets, false);
    }

    /**
     * Set the complete attribute sets for the service.  The new sets will
     * be used for all future joins.  The attribute sets at all currently
     * joined lookup services are replaced with the new sets.
     *
     * @param attrSets the attribute sets for the service
     */
    public void setAttributes(Entry[] attrSets){
	serviceItem.attributeSets = (Entry[])attrSets.clone();
	createAttributeTask(null, attrSets, SET_ATTRIBUTES); 
    }


 /** 
     * Add attribute sets for the service.  The resulting set will be used
     * for all future joins.  The attribute sets are also added at all
     * currently joined lookup services.  If checkSC is true, then throw a
     * SecurityException if any elements of attrSets are instanceof
     * ServiceControlled.
     *
     * @param attrSets the attribute sets to add
     * @param checkSC check for instanceof ServiceControlled if true
     */
    public void addAttributes(Entry[] attrSets, boolean checkSC) {
	serviceItem.attributeSets =
	    LookupAttributes.add(serviceItem.attributeSets, attrSets, checkSC);
	createAttributeTask(null, attrSets, ADD_ATTRIBUTES); 
    }



    /**  
     * Modify the current attribute sets, using the same semantics as
     * ServiceRegistration.modifyAttributes.  The resulting set will be used
     * for all future joins.  The same modifications are also made at all
     * currently joined lookup services.  If checkSC is true, then throw a
     * SecurityException if any attribute sets that would be deleted or
     * modified are instanceof ServiceControlled.
     *
     * @param attrSetTemplates the templates for matching attribute sets
     * @param attrSets the modifications to make to matching sets
     * @param checkSC check for instanceof ServiceControlled if true
     *
     * @see net.jini.core.lookup.ServiceRegistration#modifyAttributes
     */
    public void modifyAttributes(Entry[] attrSetTemplates,
               Entry[] attrSets,
               boolean checkSC)
    {
	serviceItem.attributeSets = 
	    LookupAttributes.modify(serviceItem.attributeSets, 
				    attrSetTemplates, attrSets, checkSC);
	createAttributeTask(attrSetTemplates, attrSets, MODIFY_ATTRIBUTES); 
    }

    /**
     * Cancel leases at all joined lookup services. Terminate all threads
     * within the JoinManager class.  The semantics of subsequent
     * operations on this JoinManager are undefined.
     */
    public void terminate() {
	lookupDisc.removeDiscoveryListener(grouplistener);
	locatorDisc.removeDiscoveryListener(locatorlistener);
	lookupDisc.terminate();
	locatorDisc.terminate();

	TaskThread tmpThread;
	synchronized(task) {
	    bTerminated = true;
	    tmpThread = taskThread;
	}
	if(tmpThread != null) {
	    tmpThread.interrupt();
	    try {
		tmpThread.join();
	    } catch (Exception e) {
	    }
	}

	synchronized(joinSet) {
	    for (Iterator iter = joinSet.iterator(); iter.hasNext(); ) {
		try {
		    leaseRenewalMgr.cancel(
		      ((ProxyReg)iter.next()).srvcRegistration.getLease());
		} catch (Exception ex) {
		}
	    }
	    joinSet.clear();
	}
	
    }


    private void createAttributeTask(Entry[] templs, 
				     Entry[] attrSet, int whichMethod)
    {
       AttributesTask t;
       synchronized(joinSet) {
	   if(joinSet.isEmpty())
	       return;
	   t = new AttributesTask(joinSet, templs, attrSet, whichMethod);
       }
       createTaskThread(t);
    }

    private void createTaskThread(Task t) {
       synchronized (task) {
	   if(bTerminated )
	       return;
	   task.add(t);
	   if(taskThread == null ) {
	       taskThread =  new TaskThread();
	       taskThread.start();
	   }
       }
    }

    private interface Task {
	void exec();
    }

    private class ProxyReg  {
	private class DiscLeaseListener implements LeaseListener {
  	    public void notify(LeaseRenewalEvent e ) {
		//System.out.println("**********************lease expired");
		if (e.getException() instanceof UnknownLeaseException) {
		    LeaseExpireNotifyTask t =
			new LeaseExpireNotifyTask(ProxyReg.this);
		    createTaskThread(t);
		} else 
		    fail();
	    }

	}
	public ServiceRegistrar proxy;
	public ServiceRegistration srvcRegistration = null;
	private int from;
	private DiscLeaseListener dListener = new DiscLeaseListener();
	public ProxyReg(ServiceRegistrar proxy, int from) {
	    if(proxy == null)
		throw new IllegalArgumentException(
					 "proxy canont be null");
	    this.proxy = proxy;
	    this.from = from;
	}	    

	public boolean equals(Object obj) {
	    if (obj instanceof ProxyReg){
		return proxy.equals(((ProxyReg)obj).proxy);
	    } else return false;
	}

	public int hashCode() {
	    return proxy.hashCode();
	}
	
	public void addFrom(int from) {
	    this.from |= from;
	}

	public boolean removeFrom(int from) {
	    if( from == 0 )
		    throw new RuntimeException
			("call removeFrom with bad argument");
	    this.from &= ~from;
	    return (this.from == 0);
	}
	
	
	public void register() {
	    try {
		if( proxy == null )
		    throw new RuntimeException("prox is null");
		//System.out.println("register:" + proxy.getLocator().getHost());
		   
		srvcRegistration = proxy.register(serviceItem,
						  Long.MAX_VALUE);
		Lease lease = srvcRegistration.getLease();
		leaseRenewalMgr.renewUntil(srvcRegistration.getLease(),
					   Long.MAX_VALUE, dListener);
		addJoinSet(this);
		if(serviceItem.serviceID == null ) {
		    serviceItem.serviceID = srvcRegistration.getServiceID();
		    if(callback != null )
			callback.serviceIDNotify(serviceItem.serviceID);
		}
		//System.out.println("register...succeed");
	    } catch (Exception e) {
		fail();
	    }
	    }
 
	public void addAttributes(Entry[] attSet) {
	    try {
		srvcRegistration.addAttributes(attSet);
	    } catch (Exception e) {
		fail();
	    }		    
	}

	public void modifyAttributes(Entry[] templ, Entry[] attSet) {
	    try {
		srvcRegistration.modifyAttributes(templ, attSet);
	    } catch (Exception e) {
		fail();
	    }
	}		    

	public void setAttributes(Entry[] attSet) {
	    try {
		srvcRegistration.setAttributes(attSet);
	    } catch (Exception e) {
		fail();
	    }		    
	}

	private void fail() {
	    //System.out.println("failed, call discard");
	    if((from & FROM_GROUP) != 0)
		lookupDisc.discard(proxy);

	    if((from & FROM_LOCATOR) != 0)
		locatorDisc.discard(proxy);
	}

    }

    private class AttributesTask implements Task {
	private ProxyReg[] jSet;
	private Entry[] attrSetTemplates;
	private Entry[] nsets;
	private int whichMethod = 0;

	AttributesTask(ArrayList j,  Entry[] templs, 
			      Entry[] sets, int method) 
        {
	    buildJoinSet(j, templs, sets );
	    whichMethod = method;
	}

	private void buildJoinSet(ArrayList j, Entry[] templs, Entry[] sets) {
	    jSet = new ProxyReg[j.size()];
	    j.toArray(jSet);
	    attrSetTemplates = ( templs != null ) ? 
		(Entry[]) templs.clone() : templs;
	    
	    nsets = (sets != null) ? (Entry[])sets.clone() : sets;

	}
	public void exec() {
	    for (int i=0; i<jSet.length; i++) {
		ProxyReg reg = jSet[i];
		if(!joinSet.contains(reg))
		    return;
		switch(whichMethod) {
		case ADD_ATTRIBUTES: reg.addAttributes(nsets); break;
		case MODIFY_ATTRIBUTES: 
		    reg.modifyAttributes(attrSetTemplates, nsets);break;
		case SET_ATTRIBUTES:
		    reg.setAttributes(nsets);break; 
		}
	    }
	}

    }

    private class RegisterTask implements Task {
	private ServiceRegistrar proxys[];
	private int from;
	RegisterTask(ServiceRegistrar proxys[], int from){
	    this.proxys = proxys;
	    this.from = from;
	}

	public void exec() {
	    for(int i=0; i<proxys.length; i++) {
		//if the serviceItem is a lookup service, don't register.
		if(!proxys[i].equals(serviceItem.service)) {
		    ProxyReg reg = new ProxyReg(proxys[i], from);
		    if(joinSet.contains(reg))
			reg.addFrom(from);
		    else
			reg.register();
		}
	    }
	}

    }
  
   private class LeaseExpireNotifyTask implements Task {
       protected ProxyReg reg;
       protected int from;
       LeaseExpireNotifyTask(ProxyReg reg){
	   this.reg = reg;
       }

       public void exec() {
	   if(joinSet.contains(reg))
	       reg.register();
       }

   }

    private class DiscardProxyTask implements Task {
	private ServiceRegistrar proxys[];
	private int from;
	public DiscardProxyTask(ServiceRegistrar[] proxys, int from){
	    this.proxys = proxys;
	    this.from = from;
	}

	public void exec() {
	    ArrayList drops = new ArrayList(1);
	    for(int i=0; i<proxys.length; i++) {
		ProxyReg reg = findReg(proxys[i]);
		if(reg != null ) {
		    if(reg.removeFrom(from)) {
			synchronized(joinSet) {
			    joinSet.remove(reg);
			}
			drops.add(reg);
		    }
		}
	    }
	    Iterator iter = drops.iterator();
	    while(iter.hasNext()) {
		ProxyReg reg =(ProxyReg)iter.next();
		try {
		    //@@@ call listeners's discard() in the future
		    leaseRenewalMgr.cancel(reg.srvcRegistration.getLease());
		} catch (Exception e) {
		}
	    }
	}
    }
    
    private class TaskThread extends Thread {
	public TaskThread() {
	    super("taskThread");
	    setDaemon(true);
	}

	public void run(){
	    while(!isInterrupted()) {
		Task t;
		synchronized (task) {
		    if(task.isEmpty()) {
			taskThread = null;
			return;
		    }
		    t = (Task)task.remove(0);
		}
		t.exec(); 
	    }


	}
	
    }

    private void addJoinSet(ProxyReg reg) {
	if(reg.proxy == null || reg.srvcRegistration == null)
	    throw new IllegalArgumentException(
			"bad proxy or ServiceRegistration");
	synchronized(joinSet) {
	    if(!joinSet.contains(reg)) {
		joinSet.add(reg);
	    }
	}
    }

    ProxyReg findReg(ServiceRegistrar proxy) {
	for (Iterator iter = joinSet.iterator(); iter.hasNext(); ) {
	    ProxyReg reg =(ProxyReg)iter.next();
	    if(reg.proxy.equals(proxy))
		return reg;
	}
	return null;
    }


}
