/*
 * Copyright (c) 1999 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are subject to the Sun Community Source License
 * Jini(TM) Technology Core Platform, v.1.0 (the "License"); you may not use
 * this file except in compliance with the License.  You may obtain a copy of
 * the License at http://java.sun.com/products/jini. Software distributed
 * under the License is distributed on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied.  See the License for the specific
 * language governing rights and limitations under the License.
 * 
 * CopyrightVersion v1.0_Jini
 */
package com.sun.jini.mahalo.log;

import net.jini.admin.Administrable;
import com.sun.jini.system.FileSystem;

import java.io.*;


import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Arrays;

/**
 * @author Gary Holness
 */
public class MultiLogManager
        implements LogManager, FileModes, Administrable, MultiLogManagerAdmin {
    private static final int BEGINNING = 0;
    private static final String VERSION_FILE = "Version";
    private static final String LOG_FILE = "Log.";
    private static final String TXNLOGDIR = "txnlog";
    private static final long INITIAL_LOG_COUNT = 0;
    private static final String sep = System.getProperty("file.separator");

    private static final boolean DEBUG = false;

    private LogRecovery client;
    private long logcount;
    private Map catalogue;
    private File tmpfile;
    private RandomAccessFile versionfile;
    private String directory;
    private Map files;

    private static FilenameFilter filter =
        new FilenameFilter() {
            public boolean accept(File dir, String name) {
                if (DEBUG)
                    System.out.println("MultiLogManager: recover: filter" +
                    ": name = " + name + ", matchwith = " + LOG_FILE);
                return name.startsWith(LOG_FILE);
            };
        };

    /**
     * Create a <code>MultiLogManager</code>.
     *
     * @param client who to inform during recovery.
     *
     * @param path where to store logging information.
     */
    public MultiLogManager(LogRecovery client, String path) {
	if (path == null)
	    throw new IllegalArgumentException("MultiLogManager: must use " +
						"non-null path");

	directory = path; 

	if (!directory.endsWith(sep))
	    directory = directory.concat(sep);

	if (DEBUG)
	    System.out.println("MultiLogManager: directory = " + directory);

	init(client);
    }

    /**
     * Create a <code>MultiLogManager</code> using a default
     * path for logging information storage.
     *
     * @param client who to informa during recovery
     */
    public MultiLogManager(LogRecovery client) {
	String currdir = System.getProperty("user.dir");
	directory = currdir + sep + TXNLOGDIR + sep;
	if (DEBUG)
	    System.out.println("MultiLogManager: directory = " + directory);
	init(client);
    }

    private void init(LogRecovery client) {
	if (client == null)
	    throw new IllegalArgumentException("MultiLogManager: must use " +
						"non-null recovery client");
        this.client = client;
        tmpfile = new File(directory);
 
        //If you attempt to access the Version file and
        //it does not exist, start with zero.
 
        try {
            if (DEBUG)
                System.out.println("MultiLogManager: path = " +
                                    directory + VERSION_FILE);
            if (!tmpfile.exists())
                tmpfile.mkdirs();
 
            versionfile = new RandomAccessFile(directory +  VERSION_FILE,
							    READWRITE);
 
            if (DEBUG)
                System.out.println("MultiLogManager: versionfile = " +
							versionfile);
 
            logcount = readVersion();
            writeVersion(logcount);
        } catch (IllegalArgumentException iae) {
            System.out.println("MultiLogManager: error reading Version File\n");
            iae.printStackTrace();
        } catch (IOException ioe) {
            System.out.println("MultiLogManager: error reading Version File\n");
            ioe.printStackTrace();
        } catch (SecurityException se) {
            System.out.println("MultiLogManager: error reading Version File\n");
            se.printStackTrace();
        }
 
        catalogue = Collections.synchronizedMap(new HashMap());
        files = Collections.synchronizedMap(new HashMap());
    }


    private long readVersion() {
	long val = INITIAL_LOG_COUNT;
	try {
	    versionfile.seek(BEGINNING);
	    val = versionfile.readLong();
	} catch(EOFException eofe) {
	    //versionfile is empty which means it
	    //was just created, so swallow the
	    //exception
	    if (DEBUG)
	        System.out.println("MulitLogManager: readVersion: " +
				    "initial version");
	} catch (IOException ioe) {
	    System.out.println("MultiLogManager: readVersion: " +
						ioe.getMessage());
	    ioe.printStackTrace();
	}

	if (DEBUG)
	    System.out.println("MultiLogManager: readVersion: val = " + val);
	return val;
    }

    private void writeVersion(long ver) {
	try {
	    versionfile.seek(BEGINNING);
	    versionfile.writeLong(ver);
	    if (DEBUG)
		System.out.println("MultiLogManager: writeVersion: ver = " +
								     ver);
	} catch (IOException ioe) {
	    System.out.println("MultiLogManager: writeVersion: " +
						ioe.getMessage());
	    ioe.printStackTrace();
	}
    }


    /**
     *  Factory method which returns a Log interface which
     *  can be used to create persistent records which
     *  collectively represent a system's state.
     *
     * @param cookie identifier associated with logged information.
     */
    public ClientLog logFor(long cookie) throws LogException {
	ClientLog cl = null;
	Log log = null;
	Long fileindex = null;
	Long key = new Long(cookie);

	//Maintain a mapping between cookie values and
	//fileindex.  This is done so we can name log
	//file sequentially regardless of how cookies
	//are chosen.

	fileindex = (Long)catalogue.get(key); 

	if (fileindex == null) {

	    //one person at a time can update the
	    //number which controls Log.# filenames
	    //so we don't get a race condition which
	    //would result in a non sequential
	    //numbering of log file on the file
	    //system.

	    synchronized (this) {
		fileindex = new Long(logcount);
		logcount++;
		writeVersion(logcount);
	    }

	    if (DEBUG)
	        System.out.println("MultiLogManager: logFor: fileindex = " +
					        fileindex + " key = " + key);
	    catalogue.put(key, fileindex);

	    log = new SimpleLogFile(directory + LOG_FILE +
				fileindex.longValue(), cookie,
						catalogue, files);
	    if (DEBUG)
		System.out.println("MultiLogManager: logFor: created: " +
				    directory +  LOG_FILE +
				    fileindex.longValue());
	    files.put(key, log);

	    cl = (ClientLog) log;

	    if (DEBUG)
		System.out.println("MultiLogManager: logFor: cl = " + cl);
	} else {
	    if (DEBUG)
	        System.out.println("MultiLogManager: logFor: fileindex = " +
					        fileindex + " key = " + key);

	    cl = (ClientLog) files.get(key);

	    if (DEBUG)
		System.out.println("MultiLogManager: logFor: using ClientLog " +
				    "for cookie(" + cookie + ")");
	}
	return cl;
    }

    /**
     *  Consumes the log file and re-constructs a system's
     *  state.
     */
    public void recover() throws LogException {
	Log log = null;
	File tmpfile = null;
	String [] filenames = null;
	logcount = readVersion();
	try {
	    tmpfile = new File(directory);
	    filenames = tmpfile.list(filter);

	    if (filenames.length == 0)
	      return;

	    Arrays.sort(filenames);

	    for (int i = 0; i < filenames.length; i++ ) {
	        log = new SimpleLogFile(directory +  filenames[i],
						    catalogue, files);
		if (DEBUG)
		    System.out.println("MultiLogManager: recover: " +
					"processing: " + directory +
							filenames[i]);
		log.recover(client);

		//Add the file first so that there will not be
		//any dangling mappings
                files.put(new Long(log.cookie()),log);
		catalogue.put(new Long(log.cookie()),new Long(i));
	    }
	} catch (LogException le) {
	    System.out.println("MultiLogManager: recover: unable to recover " +
								    "state");
	    le.printStackTrace();
	} catch (NullPointerException npe) {
	    System.out.println("MultiLogManager: recover: unable to recover " +
								    "state");
	    npe.printStackTrace();
	} catch (ClassCastException cce) {
	    System.out.println("MultiLogManager: recover: unable to recover " +
								    "state");
	    cce.printStackTrace();
	} catch (SecurityException se) {
	    System.out.println("MultiLogManager: recover: unable to recover " +
								    "state");
	    se.printStackTrace();
	}
    }


    /**
     * Retrieves the administration interface for the
     * <code>MultiLogManager</code>
     *
     */
    public Object getAdmin() {
	return (MultiLogManagerAdmin)this;
    }


    /**
     * Clean up all <code>LogFile</code> objects on behalf of caller.
     *
     * @see com.sun.jini.admin.DestroyAdmin
     * @see com.sun.jini.system.FileSystem
     */
    public void destroy() {
        Log log = null;
        File tmpfile = null;
        String [] filenames = null;
        logcount = readVersion();
	Object [] fls = null;

	//If MultiLogManager has been instantiated and
	//filled will LogFile associations (either through
	//explicit creation or recovery), then we destroy
	//by invalidating all known LogFiles.  Otherwise,
	//we search for LogFiles and, then, invalidate
	//them.

	if (files.size() > 0) {
	    fls = files.values().toArray();
	    for (int i = 0; i < fls.length; i++) {
		try {
		    ((SimpleLogFile)fls[i]).invalidate();
                } catch (LogException le) {
                    System.out.println("MultiLogManager: recover: unable " +
                                                        "to recover state");
                    le.printStackTrace();
                }
	    }
	    files.clear();
	    catalogue.clear();
	} else {
            try {
                tmpfile = new File(directory);
                filenames = tmpfile.list(filter);
 
                if (filenames.length == 0)
                  return;
 
                Arrays.sort(filenames);
 
                for (int i = 0; i < filenames.length; i++ ) {
	            try { 
                        log = new SimpleLogFile(directory +  filenames[i],
		    			    	        catalogue, files);
                        if (DEBUG)
                            System.out.println("MultiLogManager: recover: " +
                                            "processing: " + directory +
                                                        filenames[i]);
                        log.invalidate();
        	    } catch (LogException le) {
            	        System.out.println("MultiLogManager: recover: unable " +
					    		    "to recover state");
            	        le.printStackTrace();
        	    }
                }
            } catch (NullPointerException npe) {
                System.out.println("MultiLogManager: recover: unable to " +
				    			"recover state");
                npe.printStackTrace();
            } catch (ClassCastException cce) {
                System.out.println("MultiLogManager: recover: unable to " +
				    			    "recover state");
                cce.printStackTrace();
            } catch (SecurityException se) {
                System.out.println("MultiLogManager: recover: unable to " +
				    			"recover state");
                se.printStackTrace();
            }
	}


	try {
	    if (versionfile!= null)
	        versionfile.close();
	} catch (IOException e) {
	    System.err.println("MultiLogManager: destroy: " +
				"destroying version file in " + directory);
	    e.printStackTrace();
	    System.err.println("MultiLogManager: destroy: continuing");
	}

	//Next, clean up anything that might be left behind

	try {
            FileSystem.destroy(new File(directory), true);
	} catch (IOException e) {
            System.err.println("MultiLogManager: destroy: " +
				    "destroying " + directory);
            e.printStackTrace();
            System.err.println("MultiLogManager: destroy: continuing");
	}
    }
}
