/*
 * Copyright (c) 1999 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are subject to the Sun Community Source License
 * Jini(TM) Technology Core Platform, v.1.0 (the "License"); you may not use
 * this file except in compliance with the License.  You may obtain a copy of
 * the License at http://java.sun.com/products/jini. Software distributed
 * under the License is distributed on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied.  See the License for the specific
 * language governing rights and limitations under the License.
 * 
 * CopyrightVersion v1.0_Jini
 */
package com.sun.jini.mahout.binder;

import java.util.List;

import java.io.File;
import java.io.Serializable;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.FileOutputStream;
import java.io.ObjectOutputStream;
import java.io.FileInputStream;
import java.io.ObjectInputStream;
import java.io.FileNotFoundException;
import java.rmi.MarshalledObject;

import net.jini.core.discovery.LookupLocator;
import net.jini.core.lookup.ServiceID;
import net.jini.core.entry.Entry;

import com.sun.jini.reliableLog.LogHandler;
import com.sun.jini.reliableLog.ReliableLog;
import com.sun.jini.lookup.ServiceIDListener;
import com.sun.jini.lookup.JoinManager;

/**
 * <code>JoinAdminActivationState</code> provides an implementation of
 * <code>JoinAdminState</code> that persists its state to disk, a
 * convent way of setting the state's log up, and an easy way to store
 * the state as part of an
 * <code>java.rmi.activation.ActivationDesc</code>.
 * <p>
 *
 * During service setup a <code>JoinAdminActivationState</code>
 * is created and stored as the activation state for the
 * <code>RemoteJoinAdminImpl</code>.  Once the service has been
 * exported and its initial attributes are determined
 * <code>populate()</code> is called to initialize the
 * <code>JoinAdminState</code> persistent's state.
 * <p>
 *
 * When the <code>RemoteJoinAdminImpl</code> is activated it
 * un-marshals the <code>JoinAdminActivationState</code> and calls
 * <code>unpackJoinAdminState()</code> to get the necessary
 * <code>JoinAdminState</code>. 
 *
 * @see JoinAdminState
 * @see RemoteJoinAdminImpl
 * @see JoinAdminActivationState#populate
 * @see JoinAdminActivationState#unpackJoinAdminState */

// Log structure
//
// The log for file-persistent remote join admin has several different files
// all resident in a sub-dir of the parent's log dir.
//
// The service's proxy is stored as a serialized object in a file named
// SVR_FILE_NAME.  It is the only item in the file.  The file is
// created when remote join admin is first created.  It is never changed.
//
// Reliable log is used to store the service's lookup groups,
// locators, and attributes.  The incremental logging feature is never
// used, instead a new snap shot is taken each time this content
// changes (justification being that it does not change that often so
// the cost is low).  The groups and locators are written as arrays of
// objects.  The attributes are packaged individually as Marshaled
// Objects to preserve their codebases and so the whole lot does not
// get lost by the loss of one code base.  A single int is written
// before the attributes indicating how many attribute their.  The
// reliable log is created when the remote join admin is first created.
//
// The service id is stored as a serialized object in the file named
// SVR_ID_FILE_NAME.  It is the only item in the file.  The file is
// created when the the JoinManager calls-back serviceIDNotify() the
// first time.  It is never changed.  If the file is not present when
// the the JoinManager is created without a service id.
//
// Any needed data for binding into the registry is stored in the file
// named REG_FILE_NAME.  It contains a serialized string and a int.
// The file is created if enableRegistryBinding() is called.  Once
// created it is never changed

class JoinAdminActivationState extends LogHandler implements Serializable {
    // We only extend LogHandler so we can setup the log, we don't
    // implement the methods required to update or recover the log

    static final long serialVersionUID = -1294827489882092924L;

    /** 
     * Directory we are persisting our data to 
     * @serial
     */
    private File logDir;

    /** 
     * The service's lookup attributes 
     * @serial
     */
    private Entry[] attributes;

    /** Name of file to store the service object in. */
    final static private String SVR_FILE_NAME = "service_object";

    /** Name of file to store the registry binding data in. */
    final static private String REG_FILE_NAME = "registry_data";

    /** Name of file to store to store the service's lookup ID in */
    final static private String SVR_ID_FILE_NAME = "service_id";

    /** Use an empty array of Entries in a bunch of places, create just
        one */
    final static private Entry[] eptyEntryArray = new Entry[0];

    /**
     * Create a <code>JoinAdminActivationState</code>.
     *
     * @param svcStoreDirName Directory where the state for the associated
     * <code>JoinAdminState</code> should be persisted.
     *
     * @param prefix String that should be the prepended to the name
     * of any files/directories placed in
     * <code>svcStoreDirName</code>.  May be <code>null</code>.
     */
    JoinAdminActivationState(String svcStoreDirName, String prefix)
	throws IOException
    {
	if (prefix == null)
	    prefix = "";

	File svrStoreDir = new File(svcStoreDirName);
	logDir = new File(svrStoreDir, prefix + "JoinAdminLog");
	if (logDir.exists() && logDir.isDirectory())
	    return;

	if (!logDir.exists()) {
	    if (logDir.mkdirs())
		return;
	    else
		throw new IOException("Can't create JoinAdmin log directory:" +
				      logDir);
	}

	throw new IOException("Can't create JoinAdmin log directory, " +
			      "conflicting file:" + logDir);

    }

    /**
     * Utility method to write out an array of entities to an
     * <code>ObjectOutputStream</code>.  Can be recovered by a call
     * to <code>readAttributes()</code>
     * <p>
     * Packages each attribute in its own <code>MarshalledObject</code> so 
     * a bad codebase on an attribute class will not corrupt the whole array.
     * @see JoinAdminActivationState#readAttributes
     */
    static private void writeAttributes(Entry[] attributes,
					ObjectOutputStream out) 
	throws IOException
    {
	// Need to package each attribute in its own marshaled object,
	// this makes sure that the attribute's code base is preserved
	// and when we unpack to discard attributes who's codebase
	// has been lost without throwing away those we can still deal with. 
	
	out.writeInt(attributes.length);
	for (int i=0; i<attributes.length; i++) 
	    out.writeObject(new MarshalledObject(attributes[i]));
    }

    /**
     * Utility method to read in an array of entities from a
     * <code>ObjectInputStream</code>.  Array should have been written 
     * by a call to <code>writeAttributes()</code>
     * <p>
     *
     * Will try and recover as many attributes as possible.
     * Attributes which can't be recovered won't be returned but they
     * will remain in the log.
     *
     * @see JoinAdminActivationState#writeAttributes 
     */
    static private Entry[] readAttributes(ObjectInputStream in) 
	throws IOException, ClassNotFoundException
    {
	final List entries = new java.util.LinkedList();
	final int objectCount = in.readInt();
	for (int i=0; i<objectCount; i++) {
	    try {
		MarshalledObject mo = (MarshalledObject)in.readObject();
		entries.add(mo.get());
	    } catch (IOException e) {
		System.err.println("Exception while recovering attribute:" +
				   e.getMessage());
	    } catch (ClassNotFoundException e) {
		System.err.println("Exception while recovering attribute:" +
				   e.getMessage());
	    }
	}
	return (Entry[])entries.toArray(eptyEntryArray);
    }

    /**
     * Initialize the log of the associated
     * <code>JoinAdminState</code> with the specified service and
     * lookup attributes.  Nether argument may be <code>null</code>
     * (attributes can be a zero length array.)
     */
    void populate(Object service, Entry[] attributes) throws IOException
    {
	FileOutputStream ostream =
	    new FileOutputStream(new File(logDir, SVR_FILE_NAME));
	ObjectOutputStream oostream = new ObjectOutputStream(ostream);
	oostream.writeObject(service);
	oostream.flush();
	oostream.close();
	
	// Create a reliable log for the attributes
	this.attributes = attributes;
	ReliableLog log = new ReliableLog(logDir.getAbsolutePath(), this);
	log.snapshot();
	log.close();
    }

    /**
     * Return the associated <code>JoinAdminState</code>.  Should only
     * be called once
     */
    JoinAdminState unpackJoinAdminState() throws IOException,
	ClassNotFoundException
    {
	return new FileJoinAdminState(logDir);
    }

    // ************************************************************
    // Methods required by LogHandler

    // Inherit doc comment from super interface
    public void snapshot(OutputStream out) throws IOException {
	ObjectOutputStream oostream = new ObjectOutputStream(out);
	writeAttributes(attributes, oostream);
	oostream.writeObject(new String[0]);
	oostream.writeObject(new LookupLocator[0]);
    }

    /**
     * This method always throws
     * <code>UnsupportedOperationException</code> since
     * <code>JoinAdminActivationState</code> should never recover a
     * log.
     */
    public void recover(InputStream in) {
	// Never recover
	throw new UnsupportedOperationException("JoinAdminActivationState:" +
	    "Recovering log, this should not happen");
    }

    /**
     * This method always throws
     * <code>UnsupportedOperationException</code> since
     * <code>JoinAdminActivationState</code> should never update a
     * log.
     */
    public void applyUpdate(Object update) {
	// Never update
	throw new UnsupportedOperationException("JoinAdminActivationState:" +
	    "Updating log, this should not happen");
    }

    /**
     * Implementation of JoinAdminState that persists its state to disk
     */
    static class FileJoinAdminState extends LogHandler
	implements JoinAdminState, ServiceIDListener
    {
	/** Directory we are persisting our data to */
	private File logDir;

	
	/** 
	 * <code>JoinManager</code> we created for the
	 * <code>RemoteJoinAdminImpl</code>. Also use it for the in core
	 * rep of the attributes, lookup locators, and lookup groups.
	 */
	private JoinManager lookupMgr;

	/**
	 * In core rep of the service's lookup service ID
	 */
	private ServiceID serviceID;

	/**
	 * In core rep of the service
	 */
	private Object  service;

	/**
	 * Pass through to get attribute array from the log recovery method 
	 * to the code that creates the <code>JoinManger</code>
	 */
	private Entry[] attributes;

	/**
	 * Pass through to get lookup locator array from the log
	 * recovery method to the code that creates the
	 * <code>JoinManger</code> */
	private LookupLocator locators[];

	/**
	 * Pass through to get lookup group array from the log
	 * recovery method to the code that creates the
	 * <code>JoinManger</code> 
	 */
	private String groups[];

	/**
	 * Log we are using to persist attributes, lookup groups, and
	 * lookup locators to disk
	 */
	private ReliableLog log;

	/** True if binding into registry has been enabled */
	private boolean registryEnabled = false;
	
	/** 
	 * In core rep of what port to look for the registry on, only
	 * valid if <code>registryEnabled</code> is true.
	 */
	private int port;

	/** 
	 * In core rep of what name to bind the service under in the
	 * registry.  Only valid if <code>registryEnabled</code> is
	 * true.
	 */
	private String name;

	/**
	 * Create a <code>FileJoinAdminState</code> that will initialize
	 * its state from the designated directory.
	 */
	FileJoinAdminState(File logDir) throws IOException,
	    ClassNotFoundException
	{
	    this.logDir = logDir;
	    restoreState();
	}

	/**
	 * Utility method to close a file after an
	 * <code>IOException</code>, note we swallow any
	 * <code>IOException<code> because it is assumed the exception
	 * that triggered the call is more important.  
	 */
	private void safeClose(FileInputStream s) {
	    try {
		if (s != null)
		    s.close();
	    } catch (IOException ie) {
		// ignore, not catastrophic, if we are in this method
		// then we are already cleaning up from an IOException
	    }
	}
	
	/**
	 * Initialize the object from state in the log.
	 */
	private void restoreState() throws IOException, ClassNotFoundException
	{
	    if (!logDir.isDirectory()) {
		throw new IOException("JoinAdmin Can't find log directory");
	    }

	    // The service
	    FileInputStream svrStream = null;
	    try {
		svrStream = 
		    new FileInputStream(new File(logDir, SVR_FILE_NAME));
		final ObjectInputStream svrObjStream 
		    = new ObjectInputStream(svrStream);
		service = (Object)svrObjStream.readObject();
		svrObjStream.close();
	    } catch (IOException e) {
		safeClose(svrStream);
		throw new IOException("Service proxy corrupted on disk:"
				      + e.getLocalizedMessage());
	    }

	    // Are we binding into the registry and how?
  
	    final File regFile = new File(logDir, REG_FILE_NAME);
	    registryEnabled = regFile.exists();
	    if (isRegistryBindingEnabled()) {
		FileInputStream regStream = null;
		try {
		    regStream = new FileInputStream(regFile);
		    final ObjectInputStream regObjStream 
			= new ObjectInputStream(regStream);

		    name = (String)regObjStream.readObject();
		    port = regObjStream.readInt();

		    regObjStream.close();
		} catch (IOException e) {
		    safeClose(svrStream);
		    throw new IOException("Registry info corrupted on disk:"
		        + e.getLocalizedMessage() + "\n try deleting " +
			regFile);
		} catch (ClassCastException e) {
		    safeClose(svrStream);
		    throw new IOException("Registry info corrupted on disk:"
		        + e.getLocalizedMessage() + "\n try deleting " +
			regFile);
		}
	    }

	    // look for a service id
	    final FileInputStream idStream;
	    final File idFile = new File(logDir, SVR_ID_FILE_NAME);
	    try {
		idStream = new FileInputStream(idFile);
		final ObjectInputStream idObjStream =
		    new ObjectInputStream(idStream);
		serviceID = (ServiceID)idObjStream.readObject();
		idObjStream.close();
	    } catch (FileNotFoundException e) {
		// Proably our first time...don't sweet it
	    } catch (IOException e) {
		safeClose(svrStream);
		throw new IOException("Service ID corrupted on disk:" +
				      e.getLocalizedMessage());
	    } catch (ClassCastException e) {
		throw new IOException("Service ID corrupted on disk:" +
				      e.getLocalizedMessage());
	    }

	    // Attributes, lookup locators, and groups
	    try {
		log = new ReliableLog(logDir.getAbsolutePath(), this);
		log.recover();
	    } catch (IOException e) {
		try {
		    if (log != null)
			log.close(); 
		} catch (IOException ie) {
		    // Already going to throw an IO exception
		}
		throw new IOException("Lookup group/locator/attribute log " +
		    "is corrupted:" + e.getLocalizedMessage());
	    }
	}

	// ************************************************************
	// Methods required by JoinAdminState

	// Inherit doc comment from super interface
	public JoinManager startNewJoinManager() throws IOException {
	    if (lookupMgr == null) {
		if (serviceID == null)
		    lookupMgr =
			new JoinManager(service, attributes, groups, locators,
					this, null);
		else
		    lookupMgr =
			new JoinManager(serviceID, service, attributes, groups,
					locators, null);
	    }

	    // For now we are treating the state of the JoinManager
	    // as truth for our log, now that we have a lookup
	    // manager, force it into sync with our log
	    synchronized (log) {
		log.snapshot();
	    }
	    return lookupMgr;
	}

	// Inherit doc comment from super interface
	public Object getService() {return service;}

	// Inherit doc comment from super interface
	public ServiceID getServiceID() {return serviceID;}

	// Inherit doc comment from super interface
	public void updateLookupLocators() throws IOException {
	    synchronized (log) {
		log.snapshot();
	    }
	}

	// Inherit doc comment from super interface
	public void updateGroups() throws IOException {
	    synchronized (log) {
		log.snapshot();
	    }
	}

	// Inherit doc comment from super interface
	public void updateAttributes() throws IOException {
	    synchronized (log) {
		log.snapshot();
	    }
	}

	// Inherit doc comment from super interface
	public boolean destroy() {
	    try {
		synchronized (log) {
		    log.close();
		}
		return true;
	    } catch (IOException e) {
		return false;
	    }
	}

	// Inherit doc comment from super interface
	public void recordRegistryBindingEnable(String name, int port) 
	    throws IOException
	{
	    if (isRegistryBindingEnabled())
		throw new 
		    IllegalStateException("Registry binding already enabled");

	    FileOutputStream ostream = null;
	    try {
		ostream =
		    new FileOutputStream(new File(logDir, REG_FILE_NAME));
		final ObjectOutputStream oostream = 
		    new ObjectOutputStream(ostream);
		oostream.writeObject(name);
		oostream.writeInt(port);
		oostream.flush();
		oostream.close();
	    } catch (IOException e) {
		try {
		    if (ostream != null)
			ostream.close();
		} catch (IOException ie) {
		    // We are going to throw the original exception
		}
		throw e;
	    }
		
	    registryEnabled = true;
	    this.port = port;
	    this.name = name;
	}

	// Inherit doc comment from super interface
	public boolean isRegistryBindingEnabled() {
	    return registryEnabled;
	}

	// Inherit doc comment from super interface
	public int getPort() {
	    if (!isRegistryBindingEnabled())
		throw 
		    new IllegalStateException("Registry binding not enabled");
	    return port;
	}

	// Inherit doc comment from super interface
	public String getName() {
	    if (!isRegistryBindingEnabled())
		throw 
		    new IllegalStateException("Registry binding not enabled");
	    return name;
	}


	// ************************************************************
	// Methods required by ServiceIDListener

	// Inherit doc comment from super interface
	public void serviceIDNotify(ServiceID serviceID) {
	    this.serviceID = serviceID;

	    if (logDir == null) return;

	    try {
		FileOutputStream ostream =
		    new FileOutputStream(new File(logDir, SVR_ID_FILE_NAME));
		ObjectOutputStream oostream = new ObjectOutputStream(ostream);
		oostream.writeObject(this.serviceID);
		oostream.flush();
		oostream.close();
	    } catch (IOException e) {
		System.err.println("JoinAdmin:Can't persist service id");
		e.printStackTrace();
		System.err.println("...continuing...may not recover " +
				   "service id after restart");
	    }
	}

	// ************************************************************
	// Methods required by LogHandler

	// Inherit doc comment from super interface
	public void snapshot(OutputStream out) throws IOException {
	    ObjectOutputStream oostream = new ObjectOutputStream(out);
	    writeAttributes(lookupMgr.getAttributes(), oostream);
	    oostream.writeObject(lookupMgr.getGroups());
	    oostream.writeObject(lookupMgr.getLocators());
	}

	// Inherit doc comment from super interface
	public void recover(InputStream in) throws Exception {
	    ObjectInputStream oistream = new ObjectInputStream(in);
	    attributes = readAttributes(oistream);
	    groups = (String[])oistream.readObject();
	    locators = (LookupLocator[])oistream.readObject();
	}

	/**
	 * This method always throws
	 * <code>UnsupportedOperationException</code> since
	 * <code>FileJoinAdminState</code> should never update a
	 * log.
	 */
	public void applyUpdate(Object update) throws Exception {
	    throw new UnsupportedOperationException(
                "JoinAdminActivationState.FileJoinAdminState:Updating log" +
		", this should not happen");
	}
    }
}
