/*
 * Copyright (c) 1999 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are subject to the Sun Community Source License
 * Jini(TM) Technology Core Platform, v.1.0 (the "License"); you may not use
 * this file except in compliance with the License.  You may obtain a copy of
 * the License at http://java.sun.com/products/jini. Software distributed
 * under the License is distributed on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied.  See the License for the specific
 * language governing rights and limitations under the License.
 * 
 * CopyrightVersion v1.0_Jini
 */
package com.sun.jini.reggie;

import net.jini.core.lookup.ServiceRegistrar;
import net.jini.lookup.DiscoveryAdmin;
import net.jini.admin.Administrable;
import java.util.Properties;
import java.util.StringTokenizer;
import java.util.ArrayList;
import java.net.URL;
import java.net.MalformedURLException;
import java.rmi.activation.ActivationGroupDesc;
import java.rmi.activation.ActivationGroupDesc.CommandEnvironment;
import java.rmi.activation.ActivationGroup;
import java.rmi.activation.ActivationGroupID;
import java.rmi.activation.ActivationDesc;
import java.rmi.activation.Activatable;
import java.rmi.activation.ActivationException;
import java.io.File;
import java.io.IOException;
import java.rmi.MarshalledObject;
import java.rmi.RemoteException;

/**
 * Public methods and command line interface for creating an activatable
 * instance of the lookup service.
 */
public class CreateLookup {

    private CreateLookup() {}

    /** Holds the multiple return values from the create methods. */
    public static class Created {
	/** The activation group id */
	public final ActivationGroupID gid;
	/** The lookup service proxy */
	public final ServiceRegistrar reg;

	/**
	 * Simple constructor.
	 *
	 * @param gid activation group id
	 * @param reg lookup service proxy
	 */
	public Created(ActivationGroupID gid, ServiceRegistrar reg) {
	    this.gid = gid;
	    this.reg = reg;
	}
    }

    /**
     * Create an activatable instance of the lookup service and get a proxy
     * back.  The lookup service is initially a member of no groups.  Note
     * that when the lookup service is later destroyed, it will destroy its
     * activation group, regardless of whether or not there are other
     * activatable objects still registered in the same group.
     *
     * @param gid id of the activation group to use
     * @param logDirname directory in which the lookup service's state will
     *                   be recorded for persistence
     */
    public static ServiceRegistrar create(ActivationGroupID gid,
					  String logDirname)
	throws ActivationException, IOException
    {
	ActivationDesc desc =
	    new ActivationDesc(gid, "com.sun.jini.reggie.RegistrarImpl",
			       null, new MarshalledObject(logDirname),
			       true);
	Registrar reg = (Registrar)Activatable.register(desc);
	return new RegistrarProxy(reg, reg.getServiceID());
    }

    /**
     * Create an activation group suitable for a lookup service.
     *
     * @param classpath classpath for the lookup service
     * @param codebase codebase for the lookup service
     * @param policy security policy file
     * @param props properties for the activation group, or null
     * @param cmdpath command program for the activation group, or null
     * @param argv command options for the activation group, or null
     */
    public static ActivationGroupID createGroup(String classpath,
						String codebase,
						String policy,
						Properties props,
						String cmdpath,
						String[] argv)
	throws ActivationException, RemoteException
    {
	if (props == null)
	    props = new Properties();
	else
	    props = (Properties)props.clone();
	props.put("java.security.policy", policy);
	props.put("java.rmi.server.codebase", codebase);
	if (argv == null) {
	    argv = new String[]{"-cp", classpath};
	} else {
	    String[] nargv = new String[argv.length + 2];
	    System.arraycopy(argv, 0, nargv, 0, argv.length);
	    nargv[argv.length] = "-cp";
	    nargv[argv.length + 1] = classpath;
	    argv = nargv;
	}
	CommandEnvironment cmd = new CommandEnvironment(cmdpath, argv);
	return ActivationGroup.getSystem().registerGroup(
					 new ActivationGroupDesc(props, cmd));
    }

    /**
     * Create an activatable instance of the lookup service in a new
     * activation group and get a proxy back.  The lookup service is
     * initially a member of no groups.  Note that when the lookup service
     * is later destroyed, it will destroy its activation group, regardless
     * of whether or not there are other activatable objects still registered
     * in the same group.
     *
     * @param classpath classpath for the lookup service
     * @param codebase codebase for the lookup service
     * @param policy security policy file
     * @param dbdir directory in which the lookup service's state will be
     *              recorded for persistence
     */
    public static Created create(String classpath,
				 String codebase,
				 String policy,
				 String dbdir)
	throws ActivationException, IOException
    {
	return create(classpath, codebase, policy, dbdir, null, null, null);
    }

    /**
     * Create an activatable instance of the lookup service in a new
     * activation group and get a proxy back.  The lookup service is
     * initially a member of no groups.  Note that when the lookup service
     * is later destroyed, it will destroy its activation group, regardless
     * of whether or not there are other activatable objects still registered
     * in the same group.
     *
     * @param classpath classpath for the lookup service
     * @param codebase codebase for the lookup service
     * @param policy security policy file
     * @param dbdir directory in which the lookup service's state will be
     *              recorded for persistence
     * @param props properties for the activation group, or null
     * @param cmdpath command program for the activation group, or null
     * @param argv command options for the activation group, or null
     */
    public static Created create(String classpath,
				 String codebase,
				 String policy,
				 String dbdir,
				 Properties props,
				 String cmdpath,
				 String[] argv)
	throws ActivationException, IOException
    {
	ActivationGroupID gid = createGroup(classpath, codebase, policy,
					    props, cmdpath, argv);
	try {
	    return new Created(gid, create(gid, dbdir));
	} catch (Exception e) {
	    try {
		ActivationGroup.getSystem().unregisterGroup(gid);
	    } catch (Exception ee) {
	    }
	    if (e instanceof ActivationException)
		throw (ActivationException)e;
	    else if (e instanceof IOException)
		throw (IOException)e;
	    throw (RuntimeException)e;
	}
    }

    private static final int NUMARGS = 3;

    /**
     * Command line interface for creating an instance of the lookup service.
     * The command line parameters are:
     * <pre>
     * codebase policyfile dbdir [group,...] [javaprog] [javaoptions]
     * </pre>
     *
     * @param codebase    required argument containing the location with which
     *                    the lookup service's remote classes will be 
     *                    annotated; indicating the location from which those
     *                    classes should be downloaded. This argument is
     *                    passed to the activation system's group descriptor
     *                    as the java.rmi.server.codebase property so that this
     *                    property will be set in the VM created during 
     *                    activation of the lookup service. This codebase
     *                    string must be a sequence of properly formated URLs;
     *                    where each URL must end with ".jar", ".zip", or "/".

     * @param policyfile  required argument containing the name (with optional
     *                    path) of the file containing the security policies
     *                    for the lookup service. This argument is passed to 
     *                    the activation system's group descriptor as the
     *                    java.security.policy property so that this
     *                    property will be set in the VM created during
     *                    activation of the lookup service.
     * @param dbdir       required argument containing the name (with optional
     *                    path) of the directory in which the lookup service's
     *                    state will be recorded for persistence
     * @param group(s)    optional argument(s) containing the list of lookup
     *                    services groups the lookup service to be created
     *                    should belong to. A group name equal to "none" will
     *                    result in this lookup service belonging to no 
     *                    groups; a group name of "public" will result in
     *                    this lookup service belonging to the public group.
     *                    The default is "public".
     * @param javaprog    optional argument containing the name of a file
     *                    - including the full path - that will be used as
     *                    the 'cmdpath' argument in the creation of the 
     *                    CommandEnvironment object; which specifies startup
     *                    options for the activation group. This argument can
     *                    be used to override default system properties and
     *                    to specify implementation-defined options for the
     *                    activation group. Upon activation of the lookup
     *                    service, the activation system will start the 
     *                    lookup service by executing the file represented
     *                    by this argument. Although this file can take any
     *                    form - a script, a java executable, etc. - typically
     *                    it is simply a different version of the java
     *                    command than that used to execute the main()
     *                    method of CreateLookup. In this way, one can 
     *                    activate the lookup service with a different
     *                    version of java than the current version.
     * @param javaoptions optional list of argument(s), each beginning with
     *                    a '-', that will be passed into the VM created
     *                    when the lookup service is activated. These 
     *                    arguments can be any arguments understood by
     *                    the java command (or the javaprog if the java
     *                    command was overridden). These arguments can be
     *                    used to set system properties in the lookup service.
     *                    For example, a very useful property that can be set
     *                    using these arguments is the property that turns
     *                    off the just-in-time compiler (JIT). To turn off
     *                    the JIT, simply input the string "-Djava.compiler="
     *                    as one of these arguments.
     */
    public static void main(String[] args) {
	if (args.length < NUMARGS) {
	    System.out.println("Usage: codebase policyfile dbdir [group,...] [javaprog] [javaoptions]");
	    System.out.println("    use \"none\" to specify no groups, \"public\" to specify group \"\"");
	    System.out.println("    groups defaults to \"public\"");
	    return;
	}
	String classpath = System.getProperty("java.class.path");
	String codebase = args[0];
	String policy = args[1];
	String dbdir = args[2];
	boolean error = false;
	StringTokenizer st = new StringTokenizer(codebase);
	for (int i = 0; st.hasMoreTokens(); i++) {
	    String url = st.nextToken();
	    if (!url.endsWith("/") &&
		!url.endsWith(".jar") && !url.endsWith(".zip"))
	    {
		System.out.println("codebase URL missing trailing slash: " +
				   url);
		error = true;
	    }
	    try {
		new URL(url);
	    } catch (MalformedURLException e) {
		System.out.println("codebase contains malformed URL: " + url);
		error = true;
	    }
	}
	File pf = new File(policy);
	if (!pf.exists()) {
	    System.out.println("policy file " + policy + " does not exist");
	    error = true;
	}
	File df = new File(dbdir);
	if (df.exists()) {
	    System.out.println("dbdir " + dbdir +
			       " exists, specify a different path");
	    error = true;
	}
	String[] groups = {""};
	String prog = null;
	CommandEnvironment cmd = null;
	Properties props = null;
	int i = NUMARGS;
	int j = NUMARGS;
	if (i < args.length &&
	    !args[i].startsWith("-") &&
	    args[i].indexOf(File.separatorChar) < 0)
	{
	    if (args[i].equals("none")) {
		groups = null;
	    } else {
		st = new StringTokenizer(args[i], " \t\n\r\f,");
		groups = new String[st.countTokens()];
		for (int k = 0; st.hasMoreTokens(); k++) {
		    String group = st.nextToken();
		    if (group.equals("public"))
			group = "";
		    groups[k] = group;
		}
	    }
	    i++;
	}
	if (i < args.length && !args[i].startsWith("-")) {
	    prog = args[i];
	    i++;
	    File jf = new File(prog);
	    if (!jf.exists() || jf.isDirectory() || !jf.isAbsolute()) {
		System.out.println(
			       "java program " + prog +
			       " must be absolute path to existing file");
		error = true;
	    }
	}
	for ( ; i < args.length; i++) {
	    if (!args[i].startsWith("-") &&
		(j == NUMARGS || (!args[j-1].equals("-classpath") &&
				  !args[j-1].equals("-cp"))))
	    {
		System.out.println("java option " + args[i] +
				   " must start with -");
		error = true;
	    } else if (args[i].startsWith("-D")) {
		if (props == null)
		    props = new Properties();
		int k = args[i].indexOf('=');
		if (k < 0)
		    props.put(args[i].substring(2), "");
		else
		    props.put(args[i].substring(2, k),
			      args[i].substring(k + 1));
	    } else {
		args[j++] = args[i];
	    }
	}
	if (error)
	    return;
	String[] opts = null;
	if (j > NUMARGS) {
	    opts = new String[j - NUMARGS];
	    System.arraycopy(args, NUMARGS, opts, 0, opts.length);
	}
	if (!pf.isAbsolute()) {
	    policy = pf.getAbsolutePath();
	    System.out.println("using absolute policy path: " + policy);
	}
	if (!df.isAbsolute()) {
	    dbdir = df.getAbsolutePath();
	    System.out.println("using absolute dbdir path: " + dbdir);
	}
	try {
	    ServiceRegistrar reg = create(classpath, codebase, policy,
					  dbdir, props, prog, opts).reg;
	    if (groups != null) {
		Object admin = ((Administrable)reg).getAdmin();
		((DiscoveryAdmin)admin).setMemberGroups(groups);
	    }
	} catch (Exception e) {
	    e.printStackTrace(System.out);
	}
    }
}
