/*
 * Copyright (c) 1999 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are subject to the Sun Community Source License
 * Jini(TM) Technology Core Platform, v.1.0 (the "License"); you may not use
 * this file except in compliance with the License.  You may obtain a copy of
 * the License at http://java.sun.com/products/jini. Software distributed
 * under the License is distributed on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied.  See the License for the specific
 * language governing rights and limitations under the License.
 * 
 * CopyrightVersion v1.0_Jini
 */
package com.sun.jini.reggie;

import net.jini.lookup.DiscoveryAdmin;
import net.jini.core.lookup.ServiceRegistrar;
import net.jini.admin.Administrable;
import java.rmi.RMISecurityManager;
import java.rmi.server.RemoteObject;
import java.io.File;
import java.io.IOException;
import java.util.StringTokenizer;
import java.util.ArrayList;

/**
 * Public method and command line interface for creating a transient instance
 * of the lookup service.
 */
public class TransientLookup {

    /**
     * Create a transient instance of the lookup service and get a proxy back.
     * The lookup service is initially a member of no groups.  Persistent
     * state is maintained even though the lookup service is transient.  This
     * is particularly convenient for debugging.
     *
     * @param logDirname persistent storage directory
     */
    public static ServiceRegistrar create(String logDirname) throws IOException
    {
	RegistrarImpl impl = new RegistrarImpl(logDirname);
	return new RegistrarProxy((Registrar)RemoteObject.toStub(impl),
				  impl.getServiceID());
    }

    /**
     * Command line interface for creating a transient instance of the
     * lookup service. The command line parameters are:
     * <pre>
     * codebase dbdir [group,...]
     * </pre>
     *
     * @param dbdir       required argument containing the name (with optional
     *                    path) of the directory in which the lookup service's
     *                    state will be recorded for persistence
     * @param group(s)    optional argument(s) containing the list of lookup
     *                    services groups the lookup service to be created
     *                    should belong to. A group name equal to "none" will
     *                    result in this lookup service belonging to no 
     *                    groups; a group name of "public" will result in
     *                    this lookup service belonging to the public group.
     *                    The default is "public".
     */
    public static void main(String[] args) {
	if (args.length < 1 || args.length > 2) {
	    System.out.println("TransientLookup dbdir [group,...]");
	    System.out.println("    use \"none\" to specify no groups, \"public\" to specify group \"\"");
	    System.out.println("    groups defaults to \"public\"");
	    return;
	}
	if (new File(args[0]).exists()) {
	    System.out.println("dbdir " + args[0] +
			       " exists, specify a different path");
	    return;
	}
	String[] groups = {""};
	if (args.length > 1) {
	    if (args[1].equals("none")) {
		groups = null;
	    } else {
		StringTokenizer st =
		    new StringTokenizer(args[1], " \t\n\r\f,");
		groups = new String[st.countTokens()];
		for (int i = 0; st.hasMoreTokens(); i++) {
		    String group = st.nextToken();
		    if (group.equals("public"))
			group = "";
		    groups[i] = group;
		}
	    }
	}
	System.setSecurityManager(new RMISecurityManager());
	try {
	    ServiceRegistrar reg = create(args[0]);
	    if (groups != null) {
		Object admin = ((Administrable)reg).getAdmin();
		((DiscoveryAdmin)admin).setMemberGroups(groups);
	    }
	} catch (Exception e) {
	    e.printStackTrace(System.out);
	}
    }
}
