/*
 * Copyright (c) 1999 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are subject to the Sun Community Source License
 * Jini(TM) Technology Core Platform, v.1.0 (the "License"); you may not use
 * this file except in compliance with the License.  You may obtain a copy of
 * the License at http://java.sun.com/products/jini. Software distributed
 * under the License is distributed on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied.  See the License for the specific
 * language governing rights and limitations under the License.
 * 
 * CopyrightVersion v1.0_Jini
 */
package com.sun.jini.tool;

import java.net.ServerSocket;
import java.net.Socket;
import java.io.IOException;
import java.io.File;
import java.io.FileInputStream;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.InputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.FilePermission;
import java.util.HashMap;
import java.util.ArrayList;
import java.util.StringTokenizer;
import java.util.jar.JarFile;
import java.util.jar.JarEntry;
import java.util.jar.Manifest;
import java.util.jar.Attributes;

/** A simple HTTP server, for serving up jar and class files. */
public class ClassServer extends Thread {
    /** Default HTTP port */
    private static int DEFAULT_PORT = 8080;
    /** Default directory to serve files from on non-Windows OS */
    private static String DEFAULT_DIR = "/vob/jive/jars";
    /** Default directory to serve files from on Windows */
    private static String DEFAULT_WIN_DIR = "J:";

    /** Server socket to accept connections on */
    private ServerSocket server;
    /** Directory to serve from */
    private String dir;
    /** Map from String (jar root) to JarFile[] (jar class path) */
    private HashMap map;
    /** Verbosity flag */
    private boolean verbose;
    /** Read permission on dir and all subdirs */
    private FilePermission perm;

    /**
     * Construct a server.  Use the start method to run it.
     *
     * @param port the port to use
     * @param dir the directory to serve files from
     * @param trees true if files within jar files should be served up
     * @param verbose true if downloads should be logged
     */
    public ClassServer(int port, String dir, boolean trees, boolean verbose)
	throws IOException
    {
	if (!dir.endsWith(File.separator))
	    dir = dir + File.separatorChar;
	this.dir = dir;
	this.verbose = verbose;
	server = new ServerSocket(port);
	perm = new FilePermission(dir + '-', "read");
	if (!trees)
	    return;
	map = new HashMap();
	String[] files = new File(dir).list();
	for (int i = 0; i < files.length; i++) {
	    String jar = files[i];
	    if (!jar.endsWith(".jar"))
		continue;
	    ArrayList jfs = new ArrayList(1);
	    addJar(jar, jfs, dir);
	    map.put(jar.substring(0, jar.length() - 4),
		    jfs.toArray(new JarFile[jfs.size()]));
	}
    }

    /** Add transitive Class-Path jars to jfs. */
    private void addJar(String jar, ArrayList jfs, String dir)
	throws IOException
    {
	if (verbose)
	    System.out.println(dir + jar);
	JarFile jf = new JarFile(dir + jar);
	jfs.add(jf);
	Manifest man = jf.getManifest();
	if (man == null)
	    return;
	Attributes attrs = man.getMainAttributes();
	if (attrs == null)
	    return;
	String val = attrs.getValue(Attributes.Name.CLASS_PATH);
	if (val == null)
	    return;
	dir = dir + jar.substring(0, jar.lastIndexOf(File.separatorChar) + 1);
	StringTokenizer st = new StringTokenizer(val);
	while (st.hasMoreTokens()) {
	    addJar(st.nextToken().replace('/', File.separatorChar), jfs, dir);
	}
    }

    /** Just keep looping, spawning a new thread for each incoming request. */
    public void run() {
	try {
	    while (true) {
		new Task(server.accept()).start();
	    }
	} catch (IOException e) {
	    e.printStackTrace();
	}
    }

    /** Simple task thread */
    private class Task extends Thread {
	/** Socket for the incoming request */
	private Socket sock;

	/** Simple constructor */
	public Task(Socket sock) {
	    this.sock = sock;
	}

	/** Read specified number of bytes and always close the stream. */
	private byte[] getBytes(InputStream in, long length)
	    throws IOException
	{
	    DataInputStream din = new DataInputStream(in);
	    byte[] bytes = new byte[(int)length];
	    try {
		din.readFully(bytes);
	    } finally {
		din.close();
	    }
	    return bytes;
	}

	/** Read the GET request and return the bytes of the request file. */
	private byte[] getBytes() throws IOException {
	    BufferedReader in =
		new BufferedReader(new InputStreamReader(
						      sock.getInputStream()));
	    String line = in.readLine();
	    if (!line.startsWith("GET /"))
		throw new IOException();
	    String path = line.substring(5);
	    int i = path.indexOf(' ');
	    if (i > 0)
		path = path.substring(0, i);
	    do {
		line = in.readLine();
	    } while (line.length() > 0 &&
		     line.charAt(0) != '\r' && line.charAt(0) != '\n');
	    if (verbose)
		System.out.println(path + " from " +
				   sock.getInetAddress().getHostName() +
				   ":" + sock.getPort());
	    i = path.indexOf('/');
	    if (i < 0 || map == null) {
		if (map == null || path.endsWith(".jar")) {
		    File f = new File(dir +
				      path.replace('/', File.separatorChar));
		    if (perm.implies(new FilePermission(f.getPath(), "read")))
			return getBytes(new FileInputStream(f), f.length());
		}
		throw new IOException();
	    }
	    String jar = path.substring(0, i);
	    path = path.substring(i + 1);
	    JarFile[] jfs = (JarFile[])map.get(jar);
	    if (jfs == null)
		throw new IOException();
	    for (i = 0; i < jfs.length; i++) {
		JarEntry je = jfs[i].getJarEntry(path);
		if (je == null)
		    continue;
		return getBytes(jfs[i].getInputStream(je), je.getSize());
	    }
	    throw new IOException();
	}

	/** Process the request */
	public void run() {
	    try {
		DataOutputStream out =
		    new DataOutputStream(sock.getOutputStream());
		byte[] bytes;
		try {
		    bytes = getBytes();
		} catch (Exception e) {
		    if (verbose)
			System.out.println("(not found)");
		    out.writeBytes("HTTP/1.0 404 Not found\r\n\r\n");
		    out.flush();
		    return;
		}
		out.writeBytes("HTTP/1.0 200 OK\r\n");
		out.writeBytes("Content-Length: " + bytes.length + "\r\n");
		out.writeBytes("Content-Type: application/java\r\n\r\n");
		out.write(bytes);
		out.flush();
	    } catch (Exception e) {
		if (verbose)
		    e.printStackTrace();
	    } finally {
		try {
		    sock.close();
		} catch (IOException e) {
		}
	    }
	}
    }

    /**
     * Command line interface for creating an HTTP server.
     * The command line options are:
     * <pre>
     * [-port <var>port</var>] [-dir <var>dir</var>] [-trees] [-verbose]
     * </pre>
     * The default port is 8080; the default can be overridden with
     * the -port option.  The default directory on Windows is J:
     * and the default on other systems is /vob/jive/jars; the default
     * can be overridden with the -dir option.  By default, all files
     * under this directory (including all subdirectories) are served
     * up via HTTP.  If the pathname of a file is <var>path</var> relative
     * to the top-level directory, then the file can be downloaded using
     * the URL
     * <pre>
     * http://<var>host</var>:<var>port</var>/<var>path</var>
     * </pre>
     * No caching of directory contents or file contents is performed.
     * If the -verbose option is given, then all attempts to download files
     * are output. <p>
     * 
     * The -trees option can be used to serve up individual files stored
     * within jar files in addition to the files that are served up by
     * default.  If the option is used, the server finds all jar files
     * in the top-level directory (not in subdirectories).  For each
     * jar file, if the name of the jar file is <var>name</var>.jar, then any
     * individual file named <var>file</var> within that jar file (or within
     * the jar or zip files referenced transitively in the Class-Path manifest
     * attribute, can be downloaded using a URL of the form:
     * <pre>
     * http://<var>host</var>:<var>port</var>/<var>name</var>/<var>file</var>
     * </pre>
     * When this option is used, an open file descriptor and cached
     * information is held for each jar file, for the life of the process.
     */
    public static void main(String[] args) {
	int port = DEFAULT_PORT;
	String dir = DEFAULT_DIR;
	if (File.separatorChar == '\\')
	    dir = DEFAULT_WIN_DIR;
	boolean trees = false;
	boolean verbose = false;
	for (int i = 0; i < args.length ; i++ ) {
	    String arg = args[i];
	    if (arg.equals("-port")) {
		i++;
		port = Integer.parseInt(args[i]);
	    } else if (arg.equals("-dir")) {
		i++;
		dir = args[i];
	    } else if (arg.equals("-verbose")) {
		verbose = true;
	    } else if (arg.equals("-trees")) {
		trees = true;
	    } else {
		System.out.println(
			   "[-port <port>] [-dir <dir>] [-trees] [-verbose]");
		return;
	    }
	}
	try {
	    new ClassServer(port, dir, trees, verbose).start();
	} catch (IOException e) {
	    e.printStackTrace();
	}
    }
}
